import functools
import itertools

from descriptors import cachedproperty, classonlymethod

from fate.util.sentinel import Undefined


class NestingConf:
    """Mix-in for nested collections whose members keep track of their
    position in the collection tree.

    """
    class AbortAdoption(Exception):
        pass

    def __adopt_depth__(self, name, mapping):
        depth0 = getattr(self, '__depth__', -1)

        if depth0 is None:
            # abort! we might be mid-loading.
            raise self.AbortAdoption

        depth1 = depth0 + 1

        if mapping.__depth__ is None:
            mapping.__depth__ = depth1
        else:
            assert mapping.__depth__ == depth1

    def __adopt_name__(self, name, mapping):
        if mapping.__name__ is Undefined:
            mapping.__name__ = name
        else:
            assert mapping.__name__ == name

    def __adopt_parent__(self, name, mapping):
        if mapping.__parent__ is None:
            mapping.__parent__ = self
        else:
            assert mapping.__parent__ is self

    def __adopt__(self, name, mapping):
        try:
            self.__adopt_depth__(name, mapping)
            self.__adopt_name__(name, mapping)
            self.__adopt_parent__(name, mapping)
        except self.AbortAdoption:
            return

    def __getitem__(self, key):
        value = super().__getitem__(key)

        if isinstance(value, NestedConf):
            self.__adopt__(key, value)

        return value


class NestedPathProxy:
    """Extensible and sliceable representation of a NestedConf path."""

    def __init__(self, conf, added):
        self.conf = conf
        self.added = added

    def __getitem__(self, item):
        if not isinstance(item, (slice, int)):
            raise TypeError('indices must be integers or slices not ' + item.__class__.__name__)

        if isinstance(item, int) and item < 0:
            raise ValueError('integer indices may not be negative: 0 <= x <= sys.maxsize')

        if self.conf.__name__ is Undefined:
            return None

        # reverse, include self.conf and exclude top-level collection
        parents_headless = self.conf.__parents__[:-1]
        members_path = itertools.chain(reversed(parents_headless), [self.conf])

        slce = (item.start, item.stop, item.step) if isinstance(item, slice) else (item, item + 1)
        members_sliced = itertools.islice(members_path, *slce)

        member_names = (str(conf.__name__) for conf in members_sliced)

        return '.'.join(itertools.chain(member_names, self.added))

    def __str__(self):
        return self[:] or ''


class NestedConf(NestingConf):
    """Mix-in for members of nested collections to keep track of their
    position in the collection tree.

    """
    @classonlymethod
    def nest(cls, parent, name, *args, **kwargs):
        """Construct a new NestedConf nested under `parent`."""
        instance = cls(*args, **kwargs)
        parent.__adopt__(name, instance)
        return instance

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__reset__()

    def __reset__(self):
        self.__depth__ = None
        self.__parent__ = None
        self.__name__ = Undefined

    @property
    def __parents__(self):
        ancestry = []

        node = self

        while (parent := getattr(node, '__parent__', None)):
            node = parent
            ancestry.append(node)

        return ancestry

    @property
    def __root__(self):
        return self.__parents__[-1]

    def __get_path__(self, *added):
        return NestedPathProxy(self, added)

    @property
    def __path__(self):
        return self.__get_path__()[:]


class adopt:
    """Method decorator to __adopt__ dynamically-constructed
    configuration.

    """
    def __init__(self, name, ancestry=0):
        self.name = name
        self.ancestry = ancestry

    def __call__(self, target):
        @functools.wraps(target)
        def wrapped(parent, *args, **kwargs):
            result = target(parent, *args, **kwargs)

            depth = self.ancestry
            while (depth := depth - 1) >= 0:
                parent = parent.__parent__

            parent.__adopt__(self.name, result)

            return result

        return wrapped


class at_depth:
    """Decorator and wrapper descriptor of methods of
    DecoratedNestedConf to specify the position in the nested collection
    tree at which they are to be defined.

    """
    def __init__(self, level):
        self.__level__ = level
        self.__wrapped__ = None
        self.__dict__.update(dict.fromkeys(functools.WRAPPER_ASSIGNMENTS))

    def __call__(self, desc):
        """Decorate a method."""
        if self.__wrapped__ is not None:
            raise TypeError("wrapper already set")

        functools.update_wrapper(self, desc)
        return self

    def __get__(self, instance, cls):
        """Bind wrapped method to the given instance (if any)."""
        if instance is None:
            return self

        if self.__wrapped__ is None:
            raise TypeError("nothing has been wrapped")

        if not self.__present__(instance):
            raise AttributeError("descriptor not defined for object")

        return self.__wrapped__.__get__(instance, cls)

    def __present__(self, instance):
        """Indicate whether instance should define wrapped method."""
        if isinstance(self.__level__, int):
            return self.__level__ == instance.__depth__

        (node0, node1) = (None, instance)

        for name in reversed(self.__level__.split('.')):
            if name != '*' and name != getattr(node1, '__name__', None):
                return False

            (node0, node1) = (node1, getattr(node1, '__parent__', None))
        else:
            return getattr(node0, '__depth__', None) == 0


class DecoratedNestedConf(NestedConf):

    __atdepth_members__ = ()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # determine cls's __atdepth_members__

        # cls will inherit at_depth members from bases
        atdepth_members = set(
            itertools.chain.from_iterable(
                getattr(base, '__atdepth_members__', ()) for base in cls.__bases__
            )
        )
        # but cls will ignore *any* inherited names it overrides
        atdepth_members -= cls.__dict__.keys()

        # cls's effective at_depth members are those inherited plus
        # its own at_depth members
        atdepth_members.update(
            name for (name, obj) in cls.__dict__.items()
            if isinstance(obj, at_depth)
        )

        cls.__atdepth_members__ = tuple(atdepth_members)

    @cachedproperty
    def __atdepth_hidden__(self):
        return frozenset(name for name in self.__atdepth_members__
                         if not getattr(self.__class__, name).__present__(self))

    def __dir__(self):
        # let's try just excluding those defined here
        return [name for name in super().__dir__() if name not in self.__atdepth_hidden__]



from .access import (  # noqa: F401
    AttributeChainMap,
    AttributeAccessMap,
    AttributeDict,
    AttributeProxyDict,
)

from .collection import (  # noqa: F401
    ProxyList,
)

from .lazy import (  # noqa: F401
    LazyLoadProxyMapping,
    loads,
)

from .nesting import (  # noqa: F401
    adopt,
    at_depth,
    DecoratedNestedConf,
    NestedConf,
    NestingConf,
)

from .enum import (  # noqa: F401
    CallableEnum,
    FileFormatEnum,
    NamedTupleEnum,
    SimpleEnum,
    StrEnum,
)



"""Task controller, i.e. scheduling, commands."""
import contextlib
import datetime
import os
import random
import re
import shutil
import sys
import time

import plumbum
from descriptors import cachedproperty

from fate import sched
from fate.conf import LogsDecodingError, ResultEncodingError
from fate.util.argument import ChoiceMapping
from fate.util.compat import resources
from fate.util.lazy import lazy_id
from fate.util.log import StructLogger
from fate.util.os import pid_exists

from .. import Main


class ControlCommand(Main):
    """Base command class for concrete implementations of the task
    controller command.

    """
    # default package resource path at which to look up command banner files
    banner_path = 'fate.cli.include.banner'

    # registered task schedulers
    schedulers = {
        scheduler.module_short: scheduler
        for scheduler in (sched.TieredTenancyScheduler,)
    }

    class LockConflict(Exception):
        """Indicate a conflicting (blocking) lock."""

    def __init__(self, parser):
        parser.add_argument(
            '--force',
            action='store_true',
            help='ignore run locks',
        )
        parser.add_argument(
            '--log-level',
            choices=('debug', 'info', 'warning', 'error', 'critical'),
            metavar='name',
            help="override log targets' configured levels with one of: {%(choices)s}",
        )
        parser.add_argument(
            '--no-banner',
            action='store_false',
            dest='show_banner',
            help='suppress banner',
        )
        parser.add_argument(
            '--no-prune',
            action='store_false',
            dest='prune_runs',
            help='do not remove stale run directories',
        )

        parser.add_argument(
            '--scheduler',
            action=ChoiceMapping,
            choices=self.schedulers,
            default=next(iter(self.schedulers.values())),
            help='specify task scheduler (only current option and default: %(default)s)',
        )

    @cachedproperty
    def conf_paths(self):
        """A set of the process's configuration file paths."""
        return {conf.__path__ for conf in self.conf}

    @cachedproperty
    def run_path(self):
        """The directory path at which the process stores run-time information."""
        return self.conf._prefix_.run / f'run-{os.getpid()}'

    # Regular expression with which to extract PID from run directory name
    run_path_pattern = re.compile(r'run-(\d+)')

    @cachedproperty
    def logger(self):
        """The logger through which logs are emitted under normal operation."""
        return StructLogger(self.conf.default.path_.log_, self.args.log_level)

    @property
    def safe_logger(self):
        """A flexibly-constructed logger supporting log-emission despite
        exceptions in the creation of the full-featured logger.

        This logger *may* not conform to user's log configuration but
        should otherwise perform to expectations.

        If the situation is too dire to recover -- and a logger cannot
        be constructed -- this property may be `None`.

        """
        try:
            return self.logger.safe()
        except Exception:
            return None

    @property
    def exit_stack(self):
        """Context manager handling exceptions otherwise uncaught by
        the command.

        * Configuration-related exceptions are caught to exit the
          process with an appropriate error message and exit code
          (see `exit_on_error`).

        * All other uncaught exceptions are logged and the process
          exited with an appropriate error message and appropriate exit
          code.

          Note: This latter exception-handling is disabled when display
          of tracebacks is enabled, and when a logger (of any kind)
          could not be constructed.

        """
        stack = contextlib.ExitStack()

        if not self.args.traceback and (logger := self.safe_logger):
            log_catch = logger.bind(exc_info=True).catch(
                level='CRITICAL',
                onerror=self.onerror,
                message='fatal exception of type {record[exception].type.__name__}',
            )
            stack.enter_context(log_catch)

        stack.enter_context(self.exit_on_error)

        return stack

    def onerror(self, exc):
        """Exit the process with a generic message and exit code
        reflecting the given uncaught exception.

        """
        error = str(exc)
        error_msg = error and f': {error}'
        error_name = exc.__class__.__name__

        self.parser.exit(1, f'{self.parser.prog}: fatal: {error_name}{error_msg}\n')

    def __call__(self, args, parser):
        """Execute the command."""
        with self.exit_stack:
            try:
                self.set_lock()

                try:
                    self.check_lock(args.prune_runs, args.force)
                except self.LockConflict as exc:
                    parser.error(str(exc))

                if args.show_banner:
                    self.show_banner(args.__banner_path__)

                background = getattr(args, 'background', False)

                if getattr(args, 'continual', False):
                    self.daemon(args.scheduler, background=background)
                elif not background:
                    self.serve(args.scheduler)
                else:
                    raise NotImplementedError('cannot background during non-continual service')
            finally:
                self.remove_lock()

    def daemon(self, scheduler, background=False):
        """Schedule tasks continually as a full-featured daemon."""
        if background:
            raise NotImplementedError('background')

        next_check = 0  # no initial wait

        while True:
            if (time_sleep := next_check - time.time()) > 0:
                time.sleep(time_sleep)

            run_info = self.serve(scheduler)

            next_check = run_info.next

    def serve(self, scheduler):
        """Schedule tasks that are due for execution.

        Note: This method does *not* block to wait for tasks to become
        due.

        """
        logger = self.logger.set(session=lazy_id())

        completed_tasks = scheduler(self.conf, logger)()

        for completed_task in completed_tasks:
            self.finish_task(completed_task, logger)

        logger.info(execution_count=completed_tasks.info.count,
                    scheduled_next=datetime.datetime.fromtimestamp(completed_tasks.info.next))

        return completed_tasks.info

    def finish_task(self, task, session_log):
        """Handle a ScheduledTask whose execution has completed."""
        logger = session_log.set(task=task.__name__)

        # Check for in-process exception

        if exc := task.exception():
            # Nothing more to do than to report the error
            logger.error(f'command not found on path: {exc.program}'
                         if isinstance(exc, plumbum.CommandNotFound) else str(exc))

            return

        # Pass through task's logs (subprocess stderr)

        try:
            log_records = task.logs()
        except LogsDecodingError as exc:
            log_records = exc.logs

            logger.warning(records=len(exc.errors),
                           format=exc.format,
                           error=str(exc.errors[0]),
                           msg="bad log encoding for configured format: "
                               "record(s) treated as plain text")

        for log_record in log_records:
            logger.log(*log_record)

        # Check on task subprocess exit status

        status = self.CommandStatus.status(task.returncode)

        status_record = {
            'status': str(status),
            'exitcode': task.returncode,
        }

        if status is self.CommandStatus.Error:
            logger.error(status_record, stdout=task.stdout, stderr=task.stderr)
        else:
            logger.info(status_record)

        if status is self.CommandStatus.OK:
            # Write task result (subprocess stdout)
            try:
                result_path = task.path_.result_()
            except ResultEncodingError as exc:
                result_path = exc.identifier

                logger.warning(format=exc.format,
                               error=(str(exc.errors[0]) if len(exc.errors) == 1
                                      else [str(error) for error in exc.errors]),
                               msg="bad result encoding for configured format: "
                                   "path suffix ignored")

            if result_path:
                try:
                    self.write_result(result_path, task.stdout)
                except NotADirectoryError as exc:
                    logger.error(f'cannot record result: '
                                 f'path or sub-path is not a directory: {exc.filename}')
                except PermissionError as exc:
                    logger.error(f'cannot record result: permission denied: {exc.filename}')
            else:
                logger.debug('result empty or record disabled')

    def set_lock(self):
        """Set command's run-time lock.

        The lock indicates the command's configuration via symbolic
        links such that these may dictate a lock conflict.

        """
        self.logger.debug(run_path=self.run_path, msg="setting lock")

        lock_path = self.run_path / 'conf'
        lock_path.mkdir(parents=True)

        for conf_path in self.conf_paths:
            link_path = lock_path / conf_path.name
            link_path.symlink_to(conf_path)

        self.logger.debug(lock_path=lock_path, msg="lock set")

    def check_lock(self, prune_runs, force):
        """Check for conflicting locks.

        The discovery of any lock indicating a command running with the
        same set of configuration raises a LockConflict exception,
        dependent upon the argument `force`.

        Stale run directories, belonging to non-existent processes, are
        removed, dependent upon the argument `prune_runs`.

        """
        for run_path in self.run_path.parent.glob('run-*'):
            if run_path == self.run_path:
                continue

            name_match = self.run_path_pattern.fullmatch(run_path.name)

            if not name_match:
                continue

            pid = int(name_match.group(1))

            logger = self.logger.set(run_path=run_path)

            if not pid_exists(pid):
                if prune_runs:
                    shutil.rmtree(run_path)

                    logger.debug("dead run directory: pruned")
                else:
                    logger.debug("dead run directory: ignored")

                continue

            lock_path = run_path / 'conf'

            if not lock_path.is_dir():
                logger.debug("concurrent run")
                continue

            conf_paths = {path.readlink() for path in lock_path.iterdir() if path.is_symlink()}

            if conf_paths == self.conf_paths:
                if force:
                    logger.warning("conflicting run: ignored")
                else:
                    logger.critical("conflicting run: fatality")

                    raise self.LockConflict(
                        f"conflict: PID {pid} refers to an instance that is already running "
                        f"against the same configuration (specify --force to ignore)"
                    )
            else:
                logger.debug("concurrent run")

    def remove_lock(self):
        """Remove the process's lock and run directory tree."""
        try:
            shutil.rmtree(self.run_path)
        except FileNotFoundError:
            pass

    @classmethod
    def show_banner(cls, banner_path=None):
        """Write the contents of a random command banner file to
        standard output.

        A custom package resource path may be specified in lieu of the
        default look-up path.

        """
        for path in (banner_path, cls.banner_path):
            if path and (banner_names := resources.contents(path)):
                break
        else:
            return

        banner_name = random.choice(banner_names)

        with resources.open_text(path, banner_name) as banner_text:
            sys.stdout.writelines(banner_text)


@Main.register
class Control(ControlCommand):
    """execute scheduled commands"""

    def __init__(self, parser):
        super().__init__(parser)
        parser.add_argument(
            '--continual',
            action='store_true',
            help="run task scheduler until signaled to stop",
        )
        parser.add_argument(
            '--background',
            action='store_true',
            help="run task scheduler in background (fork) during continual "
                 "operation (see --continual)",
        )

    def __call__(self, args, parser):
        if args.background and not args.continual:
            parser.error('cannot background during non-continual service')

        super().__call__(args, parser)


class Daemon(ControlCommand):
    """launch the command-execution scheduler daemon"""

    def __init__(self, parser):
        super().__init__(parser)
        parser.add_argument(
            '--foreground',
            dest='background',
            action='store_false',
            help="run task scheduler daemon in foreground (do not fork)",
        )
        parser.set_defaults(
            continual=True,
        )


class Serve(ControlCommand):
    """execute scheduled commands"""

    def __init__(self, parser):
        super().__init__(parser)
        parser.add_argument(
            '--continual',
            action='store_true',
            help="run task scheduler until signaled to stop",
        )
        parser.set_defaults(
            background=False,
        )



import schema

from fate.util.datastructure import collection


class ConfSchema(schema.Schema):
    """Configuration validation and cleaning.

    Extends `schema.Schema` to support the `ProxyCollection` base used
    by Fate configuration.

    """
    def validate(self, data, **kwargs):
        target = data.__collection__ if isinstance(data, collection.ProxyCollection) else data

        clean = super().validate(target, **kwargs)

        if target is data:
            # we're done
            return clean

        # re-cast
        return type(data)(clean)



"""Task parameterized input handling."""
import functools
import sys

from schema import Schema

from fate.util.datastructure import AttributeDict
from fate.util.format import SLoader
from fate.util.compat.types import NoneType


Infer = object()


def read(*, schema=None, empty=Infer, format='auto', file=sys.stdin):
    """Load (parameterized) input from `file` (defaulting to standard
    input).

    `schema`, if provided, is either a `Schema` or a construction
    argument for `Schema`, and is used to validate input (and to supply
    defaults).

    `empty` is a default value to treat as the deserialized input if
    `file` is empty. By default this value is inferred from the provided
    schema.

    Input is deserialized "auto-magically" according to `format`
    (defaulting to `auto`). The input serialization format may be
    specified as one of: `{}`.

    Structured input mappings are returned as instances of
    `AttributeDict`.

    """
    if not isinstance(schema, (NoneType, Schema)):
        schema = Schema(schema)

    if empty is Infer:
        if schema and isinstance(schema.schema, (dict, list)):
            empty = type(schema.schema)()
        else:
            empty = None

    if stdin := file.read():
        try:
            (params, _loader) = SLoader.autoload(stdin, format, dict_=AttributeDict)
        except SLoader.NonAutoError:
            try:
                loader = SLoader[format]
            except KeyError:
                raise ValueError(f"unsupported format: {format!r}")

            params = loader(stdin, dict_=AttributeDict)
    else:
        params = _to_attributedict(empty)

    if schema:
        return _to_attributedict(schema.validate(params))

    return params

read.__doc__ = read.__doc__.format(SLoader.__names__)


def _deep_cast_dict(cast, target):
    """Cast all instances of `dict` in the given `target` to `cast`."""
    if isinstance(target, dict):
        return cast((key, _deep_cast_dict(cast, value)) for (key, value) in target.items())

    if isinstance(target, (list, tuple)):
        return type(target)(_deep_cast_dict(cast, item) for item in target)

    return target


_to_attributedict = functools.partial(_deep_cast_dict, AttributeDict)



import errno
import os


# adapted from psutil

def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if pid <= 0:
        raise ValueError(f'invalid PID {pid}')

    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False

        if err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True

        # According to "man 2 kill" possible error values are
        # (EINVAL, EPERM, ESRCH) therefore we should never get
        # here. If we do let's be explicit in considering this
        # an error.
        raise

    return True



import collections
import functools

from .collection import ProxyDict


class AttributeAccessMap:
    """Mix-in for mapping classes such that items may additionally be
    retrieved via attribute access.

    """
    __slots__ = ()

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            pass

        try:
            getter = super().__getattr__
        except AttributeError:
            pass
        else:
            return getter(name)

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute {name!r}")


def nomap(func, exc_class):
    """Decorator base to wrap descriptor getter functions such that they
    *may not* raise `AttributeError`.

    As such, descriptors produced from wrapped functions will never fall
    back to the class's `__getattr__`.

    This is intended for use with subclasses of `AttributeAccessMap`, to
    decorate descriptor functions for which falling back to dictionary
    access -- `__getitem__` -- is problematic, (*e.g.* those involved in
    dictionary access itself).

    """
    @functools.wraps(func)
    def wrapper(self):
        try:
            return func(self)
        except AttributeError as exc:
            raise exc_class(exc)

    return wrapper


class AttributeDict(AttributeAccessMap, dict):
    """dict whose items may additionally be retrieved via attribute
    access.

    """
    __slots__ = ()


class AttributeProxyDict(AttributeAccessMap, ProxyDict):
    """mutable mapping whose items may additionally be retrieved via
    attribute access.

    """
    __slots__ = ()


#
# ...and, for example:
#
# class AttributeMapping(AttributeAccessMap, ProxyMapping):
#     """mapping whose items may additionally be retrieved via attribute access."""
#
#     __slots__ = ()
#


class AttributeChainMap(AttributeAccessMap, collections.ChainMap):
    """ChainMap whose items may additionally be retrieved via attribute
    access.

    """



import argparse
import sys


if sys.version_info < (3, 9):
    # backport BooleanOptionalAction

    class BooleanOptionalAction(argparse.Action):

        def __init__(self,
                     option_strings,
                     dest,
                     default=None,
                     type=None,
                     choices=None,
                     required=False,
                     help=None,
                     metavar=None):

            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)

                if option_string.startswith('--'):
                    option_string = '--no-' + option_string[2:]
                    _option_strings.append(option_string)

            super().__init__(
                option_strings=_option_strings,
                dest=dest,
                nargs=0,
                default=default,
                type=type,
                choices=choices,
                required=required,
                help=help,
                metavar=metavar)

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith('--no-'))

        def format_usage(self):
            return ' | '.join(self.option_strings)

elif sys.version_info < (3, 12):
    # fix BooleanOptionalAction help
    #
    # see: https://github.com/python/cpython/issues/83137
    #
    class BooleanOptionalAction(argparse.BooleanOptionalAction):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            if self.help.endswith(' (default: %(default)s)'):
                self.help = self.help[:-23]

else:
    BooleanOptionalAction = argparse.BooleanOptionalAction



"""Task result recording compatible with the Fate scheduler."""
import sys

from fate.util.format import Dumper


def write(results, /, format='json', file=sys.stdout):
    """Write task results to `file` (defaulting to standard output).

    Results are presumed to be structured, *i.e.* in the form of a
    `dict`. These will be serialized according to `format` (defaulting
    to `json`). Supported formats include: `{}`.

    Serialization may be disabled by setting `format` to `None`.

    The object provided to `file` must feature a `write` method.

    """
    try:
        dumper = str if format is None else Dumper[format]
    except KeyError:
        raise ValueError(f"unsupported format: {format!r}")

    file.write(dumper(results))

write.__doc__ = write.__doc__.format(Dumper.__names__)



import functools
import re
import os

from jinja2 import sandbox


environ = sandbox.ImmutableSandboxedEnvironment()

environ.globals.update(
    env=os.environ,
)


variable_pattern = re.compile(r'{{(?P<expr>.*?)}}')


def eval_expr(string, **context):
    expr = environ.compile_expression(string)
    return expr(**context)


def _render_match(match, **context):
    evaluation = eval_expr(match['expr'], **context)
    return str(evaluation)


def render_str(string, **context):
    replacer = functools.partial(_render_match, **context) if context else _render_match
    return variable_pattern.sub(replacer, string)


def render_complex(contents, **context):
    if isinstance(contents, list):
        return [render_complex(content, **context) for content in contents]

    if isinstance(contents, str):
        return render_str(contents, **context)

    raise TypeError("expected str or list of str not " + contents.__class__.__name__)



import argparse
import functools
import re
import textwrap

import argcmdr
import plumbum.commands.base

from .common import CommandInterface


class OneOffExecutor(CommandInterface, argcmdr.Local):
    """Base class for Fate commands that execute tasks.

    Subclasses must define `get_command` to specify the task name
    (if any) and command to execute.

    """
    redirection_command_types = (
        plumbum.commands.base.StdinRedirection,
        plumbum.commands.base.StdinDataRedirection,
    )

    @staticmethod
    def print_output(name, text):
        """Print report value text formatted appropriately for its
        length (number of lines).

        """
        if '\n' in text:
            print(f'{name}:', textwrap.indent(text, '  '), sep='\n\n')
        else:
            print(f'{name}:', text)

    @classmethod
    def print_report(cls, name, command, retcode, stdout, stderr):
        """Print a report of task command execution outcomes."""
        print('Name:', '-' if name is None else name)

        # If we're composing the command with "echo" or otherwise providing
        # stdin then let's not include that in the report:
        cmd = command.cmd if isinstance(command, cls.redirection_command_types) else command

        print('Command:', cmd)

        print()

        if retcode is None:
            print('Status: Dry Run')
            return

        print('Status:', cls.CommandStatus.status(retcode), f'(Exit code {retcode})')

        print()

        cls.print_output('Result', stdout if stdout else '-')

        if stderr:
            print()

            # make fate task logging separators -- null byte -- visual
            stderr_formatted = stderr.replace('\0', '\n\n').strip() + '\n'
            cls.print_output('Logged (standard error)', stderr_formatted)

    def __init__(self, parser):
        super().__init__(parser)

        # argcmdr built-in arguments (@)
        # fate added-in arguments (%)

        # (@) never print commands to be executed
        #     (we handle this in the report):
        parser.set_defaults(
            show_commands=False,
        )

        # (@) look up but do not actually execute commands during dry run:
        parser.add_argument(
            '-d', '--dry-run',
            action='store_false',
            dest='execute_commands',
            help="do not execute command",
        )

        # (%) copy stdout results to given path:
        parser.add_argument(
            '-o', '--stdout',
            metavar='path',
            type=argparse.FileType('w'),
            help="write command result to path",
        )

        # (%) copy stderr output to given path:
        parser.add_argument(
            '-e', '--stderr',
            metavar='path',
            type=argparse.FileType('w'),
            help="write command standard error to path",
        )

        # (@) pass command output (stdout and stderr) through (to terminal):
        parser.add_argument(
            '-p', '--print-output',
            action='store_true',
            default=False,
            dest='foreground',
            help="print command output (in addition to report)",
        )

        # (%) silence fate command execution report:
        parser.add_argument(
            '--no-report',
            action='store_false',
            dest='report',
            help="do not print command report",
        )

    def __call__(self, args):
        with self.exit_on_error:
            super().__call__(args)

    def get_command(self, args):
        """Determine task name (if any) and command to execute
        from CLI argumentation.

        Returns either just a command to execute -- plumbum
        `BaseCommand` -- or a tuple of the task name and the
        command -- `(str, BaseCommand)`.

        """
        super(argcmdr.Local, self).__call__(args)

    def prepare(self, args, parser):
        """Execute and report on task command execution."""
        try:
            command_spec = self.delegate('get_command')
        except self.local.CommandNotFound as exc:
            hint = ('\nhint: whitespace in program name suggests a misconfiguration'
                    if re.search(r'\s', exc.program) else '')
            parser.exit(127, f'{parser.prog}: error: {exc.program}: '
                             f'command not found on path{hint}\n')

        if command_spec is None:
            return

        if send := getattr(command_spec, 'send', None):
            command_args = next(command_spec)
        else:
            command_args = command_spec

        if isinstance(command_args, (list, tuple)):
            (task_name, command) = command_args
        else:
            (task_name, command) = (None, command_args)

        result = yield command

        if send:
            try:
                send(result)
            except StopIteration:
                pass
            else:
                raise ValueError("get_command() generated more than one command")

        (retcode, stdout, stderr) = result

        if args.stdout and stdout is not None:
            print(stdout, end='', file=args.stdout)
            stdout = f'[See {args.stdout.name}]'

        if args.stderr and stderr is not None:
            print(stderr, end='', file=args.stderr)
            stderr = f'[See {args.stderr.name}]'

        if args.report:
            self.print_report(task_name, command, retcode, stdout, stderr)

    # Raise no exceptions for command return codes:
    prepare.retcode = None


"""Decorator to manufacture OneOffExecutor commands from a simple
function defining method `get_command`.

"""
runcmd = functools.partial(argcmdr.cmd, base=OneOffExecutor, method_name='get_command')



#
# TODO: drop this along with Python v3.8
#
# (importlib_resources needed for e.g. files() found only in Python v3.9)
#
import importlib
import itertools

try:
    import importlib_resources
except ImportError:
    importlib_resources = None


_target_ = importlib_resources or importlib.import_module('importlib.resources')


def __dir__():
    members = itertools.chain(globals(), dir(_target_))
    return sorted(members)


def __getattr__(name):
    return getattr(_target_, name)



"""In-memory access to supported configuration files."""
from types import SimpleNamespace

from descriptors import cachedproperty

from fate.util.compat import resources
from fate.util.datastructure import (
    AttributeDict,
    AttributeAccessMap,
    LazyLoadProxyMapping,
    loads,
    NestingConf,
    SimpleEnum,
)
from fate.util.format import Loader

from ..error import (
    ConfSyntaxError,
    MultiConfError,
    NoConfError,
)


class Conf(AttributeAccessMap, NestingConf, LazyLoadProxyMapping):
    """Dictionary- and object-style access to a configuration file."""

    _Format = Loader

    class _ConfType(SimpleEnum):

        dict_ = AttributeDict
        list_ = list

    def __init__(self, name, lib, builtin, paths, filename=None, types=None, **others):
        super().__init__()

        self.__name__ = name
        self.__lib__ = lib

        self._builtin_ = builtin
        self._prefix_ = paths

        self.__filename__ = filename or f"{name}s"

        self._types_ = types

        self.__other__ = SimpleNamespace(**others)

    def __repr__(self):
        if self.__filename__ == f"{self.__name__}s":
            filename = ""
        else:
            filename = f", filename={self.__filename__!r}"

        default = repr(dict(self))

        return (f"<{self.__class__.__name__}"
                f"({self.__name__!r}, {self.__lib__!r}{filename}) "
                f"-> {default}>")

    @cachedproperty
    @loads
    def _indicator_(self):
        return self._prefix_.conf / self.__filename__

    @cachedproperty
    @loads
    def _indicator_builtin_(self):
        return resources.files(self._builtin_.path) / self.__filename__

    def _iter_builtins_(self):
        for format_ in self._Format:
            builtin_path = self._indicator_builtin_.with_suffix(format_.suffix)

            if builtin_path.is_file():
                yield builtin_path

    def _iter_paths_(self, prefix=None):
        indicator = prefix / self.__filename__ if prefix else self._indicator_

        for format_ in self._Format:
            path = indicator.with_suffix(format_.suffix)

            if path.is_file():
                yield path

    def _get_path_(self, prefix=None):
        try:
            (path, *extra) = self._iter_paths_(prefix=prefix)
        except ValueError:
            return None

        if extra:
            raise MultiConfError(path, *extra)

        return path

    @cachedproperty
    @loads
    def __path__(self):
        if path := self._get_path_():
            return path

        if self._builtin_.fallback:
            # fall back to first built-in found
            try:
                return next(self._iter_builtins_())
            except StopIteration:
                pass

        raise NoConfError("%s{%s}" % (
            self._indicator_,
            ','.join(format_.suffix for format_ in self._Format),
        ))

    @property
    @loads
    def _format_(self):
        return self.__path__.suffix[1:]

    @property
    @loads
    def _loader_(self):
        return self._Format[self._format_]

    def __getdata__(self):
        types = {
            conf_type.name: (
                self._types_ and self._types_.get(conf_type.name.rstrip('_'))
            ) or conf_type.value
            for conf_type in self._ConfType
        }

        try:
            return self._loader_(self.__path__, **types)
        except self._loader_.raises as exc:
            raise ConfSyntaxError(self._loader_.name, exc)

    @classmethod
    def fromkeys(cls, _iterable, _value=None):
        raise TypeError(f"fromkeys unsupported for type '{cls.__name__}'")



import abc
import enum
import os
import sys
from dataclasses import dataclass
from pathlib import Path

import argcomplete

from fate.util.abstract import abstractmember
from fate.util.argument import ChoiceMapping, DirAccess, FileAccess
from fate.util.compat.argument import BooleanOptionalAction
from fate.util.datastructure import StrEnum
from fate.util.format import Loader
from fate.util.term import getch
from plumbum import colors

from .. import exit_on_error, Main


class StatusSymbol(StrEnum):

    complete   = colors.bold & colors.success | '☑'  # noqa: E221
    failed     = colors.bold & colors.fatal   | '☒'  # noqa: E221
    incomplete = colors.bold & colors.info    | '☐'  # noqa: E221


class EndStatus(enum.Enum):

    complete   = (StatusSymbol.complete,   'installed')  # noqa: E221,E241
    failed     = (StatusSymbol.failed,     'failed')     # noqa: E221,E241
    incomplete = (StatusSymbol.incomplete, 'skipped')    # noqa: E221,E241

    @property
    def symbol(self):
        return self.value[0]

    @property
    def message(self):
        return self.value[1]


class TaskSymbol(StrEnum):

    comp = colors.bold | '↹'
    conf = colors.bold | '⚙'


@dataclass
class TaskPrompt(abc.ABC):

    identifier: str
    description: str
    path: Path
    exists: bool = True
    syncd: bool = True

    update_action = abstractmember()


class PathOverwrite(TaskPrompt):

    update_action = 'overwrite'


class PathUpdate(TaskPrompt):

    update_action = 'update'


class InitCommand(Main, metaclass=abc.ABCMeta):

    description = abstractmember()
    path_access = abstractmember()

    def check_access(self, path):
        try:
            self.path_access(path)
        except self.path_access.PathTypeError:
            extant_type = 'directory' if isinstance(self.path_access, FileAccess) else 'file'

            self.parser.print_usage(sys.stderr)
            self.parser.exit(71, f'{self.parser.prog}: fatal: inferred path is '
                                 f'extant {extant_type}: {path}\n'
                                 if path.exists() else
                                 f'{self.parser.prog}: fatal: inferred path is '
                                 f'inaccessible: {path}\n')
        except self.path_access.PathAccessError:
            self.parser.print_usage(sys.stderr)
            self.parser.exit(73, f'{self.parser.prog}: fatal: inferred path is '
                                 f'not read-writable: {path}\n')

    @abc.abstractmethod
    def execute(self):
        yield from ()

    @exit_on_error
    def __call__(self, args):
        executor = self.delegate('execute')

        prompt = next(executor)

        print(StatusSymbol.complete if prompt.syncd else StatusSymbol.incomplete,
              TaskSymbol[prompt.identifier],
              colors.underline & colors.dim | str(prompt.path),
              sep='  ')

        lines = 1

        if args.prompt and not prompt.syncd:
            lines += 2

            print(
                '\n_ [Y|n]',
                colors.warn[prompt.update_action] if prompt.exists else 'install',
                f'{prompt.description}?',
                end='\r',  # return
            )

            while (do_install := getch().lower()) not in 'yn\r\x03\x04':
                pass

            if do_install == '\r':
                # set empty
                do_install = 'y'
            elif do_install in '\x03\x04':
                # treat ^C and ^D as input of "n"
                do_install = 'n'

            print(colors.underline | do_install.upper())
        else:
            do_install = 'y'

        status = executor.send(do_install == 'y')

        # update status line
        print(
            f'\033[{lines}F',                                     # jump to ☐
            status.symbol,                                        # reset symbol
            '\033[{}C'.format(5 + len(str(prompt.path))),         # jump to end
            f': {prompt.description} {status.message}',           # set message
            sep='',
            end=('\n' * lines),                                   # return to bottom
        )


@Main.register
class Init(Main):
    """post-installation initializations"""

    def __init__(self, parser):
        tty_detected = sys.stdin.isatty()
        prompt_default = 'prompt' if tty_detected else 'no prompt'

        parser.add_argument(
            '--prompt',
            default=tty_detected,
            action=BooleanOptionalAction,
            help=f"prompt to confirm actions via TTY (default: {prompt_default})",
        )

    def __call__(self):
        for (index, subcommand) in enumerate(self):
            print('' if index == 0 else '\n',
                  colors.title | subcommand.description,
                  sep='',
                  end='\n\n')

            subcommand.call()


@Init.register
class Conf(InitCommand):
    """install configuration files"""

    description = 'default configuration'
    path_access = DirAccess('rw', parents=True)

    @dataclass
    class FormatPreference:

        name: str

        @property
        def suffix(self):
            return '.' + self.name

        def select(self, suffix):
            return suffix == self.suffix

        def __str__(self):
            return self.name

    def __init__(self, parser):
        formats = {loader.name: self.FormatPreference(loader.name) for loader in Loader}
        parser.add_argument(
            '--format',
            action=ChoiceMapping,
            choices=formats,
            default=formats['toml'],
            help="configuration format to prefer (default: %(default)s)",
        )

        parser.add_argument(
            'path',
            nargs='?',
            type=self.path_access,
            help="force installation to directory path (default: inferred)",
        )

    def execute(self, args, parser):
        if args.path:
            conf_prefix = args.path
        else:
            conf_prefix = self.conf._prefix_.conf

            self.check_access(conf_prefix)

        update_paths = {}

        prompt = PathOverwrite('conf', self.description, path=None)

        for conf in self.conf:
            builtins = {path.suffix: path for path in conf._iter_builtins_()}

            formats_builtin = sorted(builtins, key=args.format.select, reverse=True)

            try:
                format_builtin = formats_builtin[0]
            except IndexError:
                parser.print_usage(sys.stderr)
                parser.exit(70, f"{parser.prog}: fatal: no built-in for "
                                f"conf file '{conf.__name__}'")

            if extant := conf._get_path_(conf_prefix):
                if template := builtins.get(extant.suffix):
                    update_paths[extant] = template
                    prompt.syncd = prompt.syncd and template.read_text() == extant.read_text()
                else:
                    parser.print_usage(sys.stderr)
                    parser.exit(70, f'{parser.prog}: fatal: no built-in template for format '
                                    f'{extant.suffix[1:]} of existing conf file: {extant}')
            else:
                prompt.exists = prompt.syncd = False

                template = builtins[format_builtin]
                target_path = conf_prefix / template.name
                update_paths[target_path] = template

        pseudo_name = '{%s}' % ','.join(path.name for path in update_paths)
        prompt.path = conf_prefix / pseudo_name

        confirmed = yield prompt

        if prompt.syncd:
            yield EndStatus.complete
        elif confirmed:
            try:
                conf_prefix.mkdir(parents=True, exist_ok=True)

                for (target_path, source_path) in update_paths.items():
                    with target_path.open('wb') as t_fd, source_path.open('rb') as s_fd:
                        t_fd.writelines(s_fd)
            except OSError:
                yield EndStatus.failed
            else:
                yield EndStatus.complete
        else:
            yield EndStatus.incomplete


@Init.register
class Comp(InitCommand):
    """install shell tab-completion files"""

    description = 'shell completion'
    path_access = FileAccess('rw', parents=True)

    script_suffixes = ('', 'd', 's')

    class Shell(StrEnum):

        bash = 'bash'
        fish = 'fish'
        tcsh = 'tcsh'

        @classmethod
        def get_choices(cls):
            return sorted(str(member) for member in cls)

        @classmethod
        def get_default(cls):
            login_shell = os.getenv('SHELL')

            if not login_shell:
                return None

            shell_path = Path(login_shell)

            if not shell_path.is_file():
                return None

            shell_name = shell_path.name

            return cls.__members__.get(shell_name)

    def __init__(self, parser):
        shell_default = self.Shell.get_default()
        parser.add_argument(
            '--shell',
            choices=self.Shell.get_choices(),
            default=shell_default,
            help="shell for which to install completion "
                 + ("(default: %(default)s)" if shell_default else "(required)"),
            required=shell_default is None,
        )

        target = parser.add_mutually_exclusive_group()
        target.add_argument(
            '--system',
            default=None,
            dest='system_profile',
            action='store_true',
            help="force system-wide installation (default: inferred)",
        )
        target.add_argument(
            '--user',
            default=None,
            dest='system_profile',
            action='store_false',
            help="force user-only installation (default: inferred)",
        )
        target.add_argument(
            'path',
            nargs='?',
            type=self.path_access,
            help="force installation to file at path (default: inferred)",
        )

    def execute(self, args):
        """install shell completion"""
        # determine installation path
        if args.path:
            completions_path = args.path
        else:
            completions_path = self.conf._prefix_.completions(args.shell, args.system_profile)

            self.check_access(completions_path)

        # determine file contents
        entry_points = args.__entry_points__ or [f'{self.conf._lib_}{suffix}'
                                                 for suffix in self.script_suffixes]

        contents = argcomplete.shellcode(entry_points, shell=args.shell)

        # check file status and prepare prompt
        prompt = PathUpdate('comp', f'{args.shell} {self.description}', completions_path)

        try:
            prompt.syncd = completions_path.read_text() == contents
        except FileNotFoundError:
            prompt.exists = prompt.syncd = False
        else:
            prompt.exists = True

        # delegate prompt to controller
        confirmed = yield prompt

        # complete execution
        if prompt.syncd:
            yield EndStatus.complete
        elif confirmed:
            try:
                completions_path.parent.mkdir(parents=True,
                                              exist_ok=True)
                completions_path.write_text(contents)
            except OSError:
                yield EndStatus.failed
            else:
                yield EndStatus.complete
        else:
            yield EndStatus.incomplete



import enum
import functools


class MixedEnumMeta(enum.EnumMeta):

    # note: override only necessary prior to Python 3.12
    def __contains__(cls, obj):
        if not isinstance(obj, enum.Enum):
            for member in cls:
                if obj == member.value:
                    return True
            else:
                return False

        return super().__contains__(obj)


class MixedEnum(enum.Enum, metaclass=MixedEnumMeta):
    pass


class StrEnum(str, MixedEnum):

    def __str__(self):
        return str(self.value)


class SimpleEnumMeta(enum.EnumMeta):

    def __getitem__(cls, name):
        simple = isinstance(name, list)

        if simple:
            (ident,) = name
        else:
            ident = name

        member = super().__getitem__(ident)

        return member.value if simple else member


class SimpleEnum(enum.Enum, metaclass=SimpleEnumMeta):
    pass


class CallableEnum(enum.Enum):

    @staticmethod
    def member(func=None, *, bound=False):
        if func is None:
            return functools.partial(callable_member, bound=bound)

        return callable_member(func, bound=bound)

    @classmethod
    def membermethod(cls, func):
        return cls.member(func, bound=True)

    def __call__(self, *args, **kwargs):
        return self.value(self, *args, **kwargs)


class callable_member:

    def __init__(self, func, *, bound=False):
        functools.update_wrapper(self, func)
        self.__func__ = func
        self.__bound__ = bound

    def __call__(self, member, *args, **kwargs):
        if self.__bound__:
            return self.__func__(member, *args, **kwargs)
        else:
            return self.__func__(*args, **kwargs)


class FileFormatEnum(enum.Enum):

    @property
    def suffix(self):
        return f'.{self.name}'


def _make(cls, iterable):
    candidates = (getattr(base, '_make', None) for base in cls.mro()
                  if not issubclass(base, enum.Enum))

    try:
        make = next(filter(None, candidates))
    except StopIteration:
        pass
    else:
        return make(iterable)

    raise TypeError("no suitable namedtuple base found")


class NamedTupleEnumMeta(enum.EnumMeta):

    def __new__(metacls, cls, bases, classdict, **kwds):
        if any(hasattr(base, '_make') for base in bases):
            classdict.setdefault('_make', classmethod(_make))

        return super().__new__(metacls, cls, bases, classdict, **kwds)


class NamedTupleEnum(enum.Enum, metaclass=NamedTupleEnumMeta):
    pass



import collections
import itertools
import operator
import time

from fate.util.compat.os import get_interval

from .base import TaskProcessPool, TaskScheduler


get_tenancy = operator.attrgetter('scheduling_.tenancy')


class TieredTenancyScheduler(TaskScheduler):
    """Task scheduler grouping tasks' execution by configured "tenancy".

    The execution concurrency of tasks whose schedules coincide is
    limited. By default, tasks abide by a level of shared tenancy equal
    to a default value, which is a function of the CPU cores available
    to the execution environment. The level of shared tenancy which a
    task permits may be configured. (For example, a task requiring the
    exclusive use of all available CPUs, or a task requiring exclusive
    use of the network, might ensure that no other tasks execute
    concurrently, by configuring a scheduling tenancy of `1`.)

    The tiered tenancy scheduler queries the task schedule to collect a
    batch of scheduled tasks. This batch's lowest-tenancy tasks are
    executed first. As permitted by lower-tenancy tasks' configuration,
    tasks permitting higher tenancy *may* be executed concurrently.
    (For example: given a single task with a configured tenancy of `2`,
    one additional task *may* execute concurrently, so long as the batch
    contains additional tasks with configured tenancies of `2` or
    greater.) As a batch's lowest-tenancy tasks complete, the number of
    tasks which may execute concurrently is increased.

    During the course of task execution, should additional tasks become
    due, then subsequent batches of tasks are collected. Tasks collected
    from earlier batches -- also known as "cohorts" -- are prioritized
    over those of later batches. However, later-cohort tasks *may* be
    executed prior to earlier-cohort tasks, to ensure as many concurrent
    tasks are running as is permitted by the tenancy configurations of
    executing and to-be-executed tasks.

    """
    # Note: optimum time to wait before polling unclear.
    # For now, let's just wait a "slice":
    poll_frequency = get_interval()

    def exec_tasks(self, reset=False):
        count_completed = 0

        tasks_0 = self.collect_tasks(reset=reset)
        queue = SchedulingQueue([tasks_0])

        self.logger.debug(cohort=0, size=queue[0].size, tenancies=queue[0].infomap,
                          msg='enqueued cohort')

        while task_0 := queue.get_task():

            pool = TaskProcessPool([task_0], size=1)

            self.logger.debug(active=pool.count, msg='launched pool')

            while pool.active:
                min_tenancy = min(get_tenancy(task) for task in pool.iter_tasks())

                if min_tenancy > pool.size:
                    pool.expand(queue.tenancy_tasks(min_tenancy), size=min_tenancy)
                    self.logger.debug(tenancy=pool.size, active=pool.count, msg='expanded pool')

                if time.time() >= self.next_check:
                    tasks_1 = self.collect_tasks(reset=True)
                    queue.append(tasks_1)

                    self.logger.debug(cohort=(len(queue) - 1), size=queue[-1].size,
                                      tenancies=queue[-1].infomap, msg='enqueued cohort')

                    count_fill = pool.fill(queue.tenancy_tasks(pool.size))

                    if count_fill:
                        self.logger.debug(active=pool.count, msg='filled pool')

                time.sleep(self.poll_frequency)

                count_ready = yield from pool.iter_ready(refill=queue.tenancy_tasks(pool.size))

                if count_ready:
                    count_completed += count_ready

                    self.logger.debug(completed=count_ready,
                                      total=count_completed,
                                      active=pool.count)

        return count_completed


class TenancyGroup:
    """Iterator of tasks sharing a given tenancy."""

    def __init__(self, tenancy, tasks):
        self.tenancy = int(tenancy)
        self.tasks = tuple(tasks)
        self.index = 0

    def __repr__(self):
        return (f"<{self.__class__.__name__}: "
                f"{{index={self.index!r}, done={self.done!r}}} "
                f"(tenancy={self.tenancy!r}, tasks={list(self.tasks)!r})>")

    def __iter__(self):
        return self

    def __next__(self):
        try:
            task = self.get()
        except IndexError:
            raise StopIteration

        self.index += 1
        return task

    def get(self):
        return self.tasks[self.index]

    @property
    def size(self):
        return len(self.tasks)

    @property
    def done(self):
        return self.index >= self.size

    @property
    def remaining(self):
        return self.size - self.index

    @property
    def info(self):
        return (self.tenancy, self.size)

    @property
    def status(self):
        return (self.tenancy, self.size, self.remaining)


def enumerate_backwards(seq):
    indices = range(len(seq) - 1, -1, -1)
    return zip(indices, reversed(seq))


class SchedulingCohort:
    """Collection of tasks resulting from the same scheduling "check".

    Tasks are sorted and grouped by their configured tenancy as
    TenancyGroups.

    Groups may be iterated via method `tenancy_groups`.

    Tasks may be iterated via method `tenancy_tasks`.

    Note: `tenancy_tasks` generates tasks by iteration over
    the cohort's TenancyGroups, consuming these iterators.
    `tenancy_groups` generates TenancyGroups from the cohort, and does
    not alter these iterators on its own. (However, given `prune=True`,
    either method will remove exhausted groups from the cohort.)

    To generate only those tasks configured with at least a minimum
    tenancy, specify this value to `tenancy_tasks`, _e.g._:

        for task in cohort.tenancy_tasks(tenancy=5):
            ...

    """
    group_class = TenancyGroup

    sort_key = get_tenancy

    def __init__(self, tasks=(), prune=True):
        # reverse in order to permit direct pruning of list
        tasks_sorted = sorted(tasks, key=self.sort_key, reverse=True)
        self._groups = [self.group_class(tenancy, group)
                        for (tenancy, group) in itertools.groupby(tasks_sorted, self.sort_key)]

        self.prune = prune

        self._done = False

    def __repr__(self):
        return f"<{self.__class__.__name__}: {list(reversed(self._groups))!r}>"

    def __len__(self):
        return len(self._groups)

    @property
    def done(self):
        # once done, cache this, rather than continue to check
        if not self._done:
            self._done = all(group.done for group in self._groups)

        return self._done

    @property
    def size(self):
        return sum(group.size for group in self._groups)

    @property
    def info(self):
        return [group.info for group in reversed(self._groups)]

    @property
    def infomap(self):
        return dict(self.info)

    def tenancy_groups(self):
        for (index, group) in enumerate_backwards(self._groups):
            if not group.done:
                yield group

            if self.prune and group.done:
                del self._groups[index]

    def tenancy_tasks(self, tenancy=1):
        for group in self.tenancy_groups():
            if group.tenancy >= tenancy:
                yield from group


class SchedulingQueue:
    """Collections of tasks resulting from various scheduling "checks".

    Batches of tasks -- or "cohorts" -- may be added to the queue
    dynamically via methods "append" and "extend".

    Supports the same group-querying and task-consumption operations as
    SchedulingCohort, but which may span multiple collected cohorts.

    """
    cohort_class = SchedulingCohort

    def __init__(self, cohorts=(), prune=True):
        # note: will be reversed in order to permit pruning of deque
        self._cohorts = collections.deque()

        self.prune = prune

        self.extend(cohorts)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {list(reversed(self._cohorts))!r}>"

    def __len__(self):
        return len(self._cohorts)

    def __getitem__(self, index):
        if isinstance(index, slice):
            raise NotImplementedError(f'{self.__class__.__name__} does not support slices')

        if not isinstance(index, int):
            raise TypeError(f'{self.__class__.__name__} indices must be integers, '
                            f'not {index.__class__.__name__}')

        try:
            return self._cohorts[-1 - index]
        except IndexError:
            pass

        raise IndexError(f'{self.__class__.__name__} index out of range')

    @property
    def size(self):
        return sum(cohort.size for cohort in self._cohorts)

    def append(self, cohort):
        self._cohorts.appendleft(self.cohort_class(cohort, prune=self.prune))

    def extend(self, cohorts):
        self._cohorts.extendleft(self.cohort_class(cohort, prune=self.prune) for cohort in cohorts)

    def cohorts(self):
        for (index, cohort) in enumerate_backwards(self._cohorts):
            if not cohort.done:
                yield cohort

            if self.prune and cohort.done:
                del self._cohorts[index]

    def tenancy_groups(self):
        for cohort in self.cohorts():
            yield from cohort.tenancy_groups()

    def tenancy_tasks(self, tenancy=1):
        for cohort in self.cohorts():
            yield from cohort.tenancy_tasks(tenancy)

    def get_task(self, tenancy=1, default=None):
        return next(self.tenancy_tasks(tenancy), default)



import time


class Timing:

    class StaleTimer(ValueError):
        pass

    def __init__(self):
        self.t0 = self.t1 = self.proc0 = self.proc1 = None

    def start(self):
        if self.t0 is not None:
            raise self.StaleTimer

        self.t0 = time.time()
        self.proc0 = time.process_time()

    def stop(self):
        if self.t1 is not None:
            raise self.StaleTimer

        self.t1 = time.time()
        self.proc1 = time.process_time()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc_info):
        self.stop()

    @property
    def time_elapsed(self):
        if self.t0 is None or self.t1 is None:
            return None

        return self.t1 - self.t0

    @property
    def proc_elapsed(self):
        if self.proc0 is None or self.proc1 is None:
            return None

        return self.proc1 - self.proc0

    def __iter__(self):
        yield self.time_elapsed
        yield self.proc_elapsed

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self}>'

    def __str__(self):
        return (f'total time elapsed {self.time_elapsed} | '
                f'process time elapsed {self.proc_elapsed}')



from descriptors import cachedproperty

from .event import event_id


class LazyStr(str):

    def __new__(cls, *_args, **_kwargs):
        # ensure str ignores arguments and
        # never instantiates anything more than empty ''
        return super().__new__(cls)

    def __init__(self, func, *args, **kwargs):
        self.__func__ = func
        self.__args__ = args
        self.__kwargs__ = kwargs

        # set lazily:
        # self.__value__

    @cachedproperty
    def __value__(self):
        value = self.__func__(*self.__args__, **self.__kwargs__)

        if not isinstance(value, str):
            raise TypeError(f'{self.__class__.__name__} expected str but got {value!r}')

        return value

    @property
    def __cached__(self):
        return self.__dict__.get('__value__')

    def __repr__(self):
        args_spec = (
            ', ' + ', '.join(repr(arg) for arg in self.__args__)
        ) if self.__args__ else ''

        kwargs_spec = (
            ', ' + ', '.join(f'{key}={value!r}' for (key, value) in self.__kwargs__.items())
        ) if self.__kwargs__ else ''

        cached = self.__cached__

        cached_spec = 'UNKNOWN' if cached is None else repr(cached)

        return (f'<{self.__class__.__name__}({self.__func__!r}{args_spec}{kwargs_spec}): '
                f'{cached_spec}>')

    def __str__(self):
        return self.__value__

    # ...and the rest...

    def __add__(self, other):
        return self.__value__.__add__(other)

    def __contains__(self, item):
        return self.__value__.__contains__(item)

    def __eq__(self, other):
        return self.__value__.__eq__(other)

    def __format__(self, format_spec):
        return self.__value__.__format__(format_spec)

    def __ge__(self, other):
        return self.__value__.__ge__(other)

    def __getitem__(self, key):
        return self.__value__.__getitem__(key)

    def __gt__(self, other):
        return self.__value__.__gt__(other)

    def __iter__(self):
        return self.__value__.__iter__()

    def __le__(self, other):
        return self.__value__.__le__(other)

    def __len__(self):
        return self.__value__.__len__()

    def __lt__(self, other):
        return self.__value__.__lt__(other)

    def __mod__(self, other):
        return self.__value__.__mod__(other)

    def __mul__(self, other):
        return self.__value__.__mul__(other)

    def __ne__(self, other):
        return self.__value__.__ne__(other)

    def __rmod__(self, other):
        return self.__value__.__rmod__(other)

    def __rmul__(self, other):
        return self.__value__.__rmul__(other)

    def capitalize(self):
        return self.__value__.capitalize()

    def casefold(self):
        return self.__value__.casefold()

    def center(self, *args, **kwargs):
        return self.__value__.center(*args, **kwargs)

    def count(self, *args, **kwargs):
        return self.__value__.count(*args, **kwargs)

    def encode(self, *args, **kwargs):
        return self.__value__.encode(*args, **kwargs)

    def endswith(self, *args, **kwargs):
        return self.__value__.endswith(*args, **kwargs)

    def expandtabs(self, *args, **kwargs):
        return self.__value__.expandtabs(*args, **kwargs)

    def find(self, *args, **kwargs):
        return self.__value__.find(*args, **kwargs)

    def format(self, *args, **kwargs):
        return self.__value__.format(*args, **kwargs)

    def format_map(self, *args, **kwargs):
        return self.__value__.format_map(*args, **kwargs)

    def index(self, *args, **kwargs):
        return self.__value__.index(*args, **kwargs)

    def isalnum(self, *args, **kwargs):
        return self.__value__.isalnum(*args, **kwargs)

    def isalpha(self, *args, **kwargs):
        return self.__value__.isalpha(*args, **kwargs)

    def isascii(self, *args, **kwargs):
        return self.__value__.isascii(*args, **kwargs)

    def isdecimal(self, *args, **kwargs):
        return self.__value__.isdecimal(*args, **kwargs)

    def isdigit(self, *args, **kwargs):
        return self.__value__.isdigit(*args, **kwargs)

    def isidentifier(self, *args, **kwargs):
        return self.__value__.isidentifier(*args, **kwargs)

    def islower(self, *args, **kwargs):
        return self.__value__.islower(*args, **kwargs)

    def isnumeric(self, *args, **kwargs):
        return self.__value__.isnumeric(*args, **kwargs)

    def isprintable(self, *args, **kwargs):
        return self.__value__.isprintable(*args, **kwargs)

    def isspace(self, *args, **kwargs):
        return self.__value__.isspace(*args, **kwargs)

    def istitle(self, *args, **kwargs):
        return self.__value__.istitle(*args, **kwargs)

    def isupper(self, *args, **kwargs):
        return self.__value__.isupper(*args, **kwargs)

    def join(self, *args, **kwargs):
        return self.__value__.join(*args, **kwargs)

    def ljust(self, *args, **kwargs):
        return self.__value__.ljust(*args, **kwargs)

    def lower(self, *args, **kwargs):
        return self.__value__.lower(*args, **kwargs)

    def lstrip(self, *args, **kwargs):
        return self.__value__.lstrip(*args, **kwargs)

    def maketrans(self, *args, **kwargs):
        return self.__value__.maketrans(*args, **kwargs)

    def partition(self, *args, **kwargs):
        return self.__value__.partition(*args, **kwargs)

    def removeprefix(self, *args, **kwargs):
        return self.__value__.removeprefix(*args, **kwargs)

    def removesuffix(self, *args, **kwargs):
        return self.__value__.removesuffix(*args, **kwargs)

    def replace(self, *args, **kwargs):
        return self.__value__.replace(*args, **kwargs)

    def rfind(self, *args, **kwargs):
        return self.__value__.rfind(*args, **kwargs)

    def rindex(self, *args, **kwargs):
        return self.__value__.rindex(*args, **kwargs)

    def rjust(self, *args, **kwargs):
        return self.__value__.rjust(*args, **kwargs)

    def rpartition(self, *args, **kwargs):
        return self.__value__.rpartition(*args, **kwargs)

    def rsplit(self, *args, **kwargs):
        return self.__value__.rsplit(*args, **kwargs)

    def rstrip(self, *args, **kwargs):
        return self.__value__.rstrip(*args, **kwargs)

    def split(self, *args, **kwargs):
        return self.__value__.split(*args, **kwargs)

    def splitlines(self, *args, **kwargs):
        return self.__value__.splitlines(*args, **kwargs)

    def startswith(self, *args, **kwargs):
        return self.__value__.startswith(*args, **kwargs)

    def strip(self, *args, **kwargs):
        return self.__value__.strip(*args, **kwargs)

    def swapcase(self, *args, **kwargs):
        return self.__value__.swapcase(*args, **kwargs)

    def title(self, *args, **kwargs):
        return self.__value__.title(*args, **kwargs)

    def translate(self, *args, **kwargs):
        return self.__value__.translate(*args, **kwargs)

    def upper(self, *args, **kwargs):
        return self.__value__.upper(*args, **kwargs)

    def zfill(self, *args, **kwargs):
        return self.__value__.zfill(*args, **kwargs)


lstr = LazyStr


def lazy_id():
    return lstr(event_id)



import argcmdr
import argparse
import os.path
import sys

from fate.conf import ResultEncodingError

from .. import Main, runcmd


READABLE = argparse.FileType('r')


def path_or_text(value):
    """Guess whether to use given value as-is or to treat as a
    filesystem path (from which to read a value).

    Returns either the given value OR an `open()` file descriptor (via
    `FileType`).

    """
    if value.startswith('{') or '\n' in value:
        return value

    if value == '-' or os.path.sep in value or os.path.exists(value):
        return READABLE(value)

    return value


@Main.register
class Debug(argcmdr.Command):
    """ad-hoc execution commands"""

    @runcmd('arguments', metavar='command-arguments', nargs=argparse.REMAINDER,
            help="command arguments (optional)")
    @runcmd('command', help="program to execute")
    @runcmd('-i', '--stdin', metavar='path|text', type=path_or_text,
            help="set standard input (parameterization) for command to given "
                 "path or text (specify '-' to pass through stdin)")
    def execute(context, args, parser):
        """execute an arbitrary program as an ad-hoc task"""
        try:
            cmd = context.local[args.command][args.arguments]
        except context.local.CommandNotFound:
            parser.print_usage(sys.stderr)
            raise

        if hasattr(args.stdin, 'read'):
            return cmd < args.stdin

        if args.stdin is not None:
            return cmd << args.stdin

        return cmd

    @runcmd('task', help="name of configured task to run")
    @runcmd('-i', '--stdin', metavar='path|text', type=path_or_text,
            help="override standard input (parameterization) for task to given "
                 "path or text (specify '-' to pass through stdin) "
                 "(default: from configuration)")
    @runcmd('--record', action='store_true', help="record task result")
    def run(context, args, parser):
        """run a configured task ad-hoc"""
        try:
            spec = context.conf.task[args.task]
        except KeyError:
            parser.error(f"task not found: '{args.task}'")

        if isinstance(exec_ := spec.exec_, str):
            cmd = context.local[exec_]
        else:
            (root, *arguments) = exec_
            cmd = context.local[root][arguments]

        if hasattr(args.stdin, 'read'):
            bound = cmd < args.stdin
        elif args.stdin is not None:
            bound = cmd << args.stdin
        else:
            bound = cmd << spec.param_

        (retcode, stdout, stderr) = yield (args.task, bound)

        if args.record and retcode == 0:
            try:
                result_path = spec.path_._result_(stdout)
            except ResultEncodingError as exc:
                result_path = exc.identifier

                print(f"result does not appear to be encoded as {exc.format}:",
                      "will write to file without suffix",
                      file=sys.stderr)

            try:
                context.write_result(result_path, stdout)
            except NotADirectoryError as exc:
                print('cannot record result: path or sub-path is not a directory:',
                      exc.filename,
                      file=sys.stderr)



def _is_relative_to(path, *other):
    """Return True if the path is relative to another path or False.
    """
    try:
        path.relative_to(*other)
        return True
    except ValueError:
        return False


# added to pathlib in 3.9
def is_relative_to(path, *other):
    try:
        is_relative_to = path.is_relative_to
    except AttributeError:
        return _is_relative_to(path, *other)
    else:
        return is_relative_to(*other)



import sys
import tty
import termios


def getch(n=1, /, file=sys.stdin):
    """Get n character(s) from TTY file."""
    fd = file.fileno()
    old = termios.tcgetattr(fd)

    try:
        tty.setraw(fd)
        return file.read(n)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old)


def getche(n=1, /, infile=sys.stdin, outfile=sys.stdout, end=''):
    """Print back input from getch."""
    result = getch(n, infile)
    print(result, end=end, file=outfile)
    return result



import os.path
import sys

import argcmdr

from .common import CommandInterface


class Main(CommandInterface, argcmdr.RootCommand):
    """manage the periodic execution of commands"""

    @classmethod
    def _new_parser_(cls):
        parser = super()._new_parser_()

        # enforce program name when invoked via "python -m fate"
        if parser.prog == '__main__.py':
            command = os.path.basename(sys.executable)
            parser.prog = f'{command} -m fate'

        return parser



import os
import sys
import collections

import loguru

from fate.util.datastructure import (
    adopt,
    at_depth,
    NestedConf,
)

from ..base import Conf

from ..error import ConfTypeError, ConfValueError

from .base import (
    ConfDict,
    ConfList,
    ConfType,
)


class DefaultConf(Conf):

    @property
    def path_(self):
        try:
            return self['path']
        except KeyError:
            return DefaultConfDict.nest(self, 'path')


class DefaultConfType(ConfType):

    _dev_keys_ = frozenset(('level',
                            'file'))

    _log_keys_ = _dev_keys_ | frozenset(('rotation',
                                         'compression',
                                         'retention',
                                         'directory'))

    @at_depth('path')
    @property
    def log_(self):
        log = self.get('log', '/dev/stderr')

        if isinstance(log, (str, collections.abc.Mapping)):
            if isinstance(log, NestedConf):
                log.__reset__()

            log = DefaultConfList.nest(self, 'log', [log])

        if not isinstance(log, collections.abc.Sequence):
            raise ConfTypeError(
                f'{self.__root__.__name__} value @ {self.__path__}.log: expected string, '
                f'mapping or list of same, not value of type {log.__class__.__name__} '
                f'(in {self.__root__.__path__})'
            )

        sinks = []

        for (index, sink) in enumerate(log):
            if isinstance(sink, str):
                if os.path.exists(sink):
                    key = 'directory' if os.path.isdir(sink) else 'file'
                else:
                    key = 'file' if '.' in os.path.basename(sink) else 'directory'

                sink = self.__class__.nest(log, index, {key: sink})

            sink_file = sink.get('file', '')

            allowed_keys = self._dev_keys_ if sink_file.startswith('/dev/') else self._log_keys_

            if bad_log_keys := sink.keys() - allowed_keys:
                sink_path = sink_file or sink.get('directory', '')
                log_id_stanza = f' {sink_path}' if sink_path else ''

                raise ConfTypeError(
                    f'{self.__root__.__name__} value @ {self.__path__}.log: allowed keys for '
                    f'log path{log_id_stanza} are {set(allowed_keys)} not {bad_log_keys} '
                    f'(in {self.__root__.__path__})'
                )

            sinks.append(sink)

        return sinks

    @at_depth('path.log.*')
    @property
    def target_(self):
        if path := self.get('file'):
            if 'directory' in self:
                raise ConfTypeError(
                    f"{self.__root__.__name__} value @ {self.__path__}: log configuration may "
                    f"specify 'file' or 'directory' not both (in {self.__root__.__path__})"
                )

            if path == '/dev/stdout':
                return sys.stdout

            if path == '/dev/stderr':
                return sys.stderr

            return path

        if directory := self.get('directory'):
            return os.path.join(directory, f'{self.__lib__}.{{time}}.log')

        raise ConfTypeError(
            f"{self.__root__.__name__} value @ {self.__path__}: log configuration must "
            f"specify either a 'file' or 'directory' target (in {self.__root__.__path__})"
        )

    @at_depth('path.log.*')
    @property
    @adopt('log', ancestry=1)
    def extra_(self):
        return self.__class__(
            (key, value) for (key, value) in self.items()
            if key != 'file' and key != 'directory'
        )

    @at_depth('path.log.*')
    @property
    def level_(self):
        try:
            return loguru.logger.level(self.level)
        except ValueError:
            raise ConfValueError(
                f'{self.__root__.__name__} value @ {self.__path__}.level: '
                f'unsupported log level: {self.level!r} '
                f'(in {self.__root__.__path__})'
            )


class DefaultConfDict(DefaultConfType, ConfDict):
    """Mapping of data deserialized from defaults configuration files."""


class DefaultConfList(DefaultConfType, ConfList):
    """Defaults configuration list."""



import plumbum
from descriptors import classonlymethod

from fate.conf.error import LogsDecodingError
from fate.conf.types import TaskConfDict, TaskChainMap
from fate.util.datastructure import at_depth, adopt


class ImmediateFailure:
    """Dummy Future for tasks failing to initialize."""

    returncode = None
    stdout = None
    stderr = None

    def __init__(self, exc):
        self.exception = exc

    @staticmethod
    def poll():
        return True

    ready = poll


class ScheduledTask(TaskConfDict):
    """Task class extended for processing by the operating system."""

    @classonlymethod
    def schedule(cls, task):
        """Construct a ScheduledTask extending the specified Task."""
        self = cls(task)
        task.__parent__.__adopt__(task.__name__, self)
        return self

    def __adopt_parent__(self, name, mapping):
        if mapping.__parent__ is None:
            mapping.__parent__ = self
            return

        # we've likely taken over for existing configuration via schedule().
        # rather than insist that child is in our tree, merely check
        # that its tree looks the same.
        assert isinstance(mapping.__parent__, TaskConfDict)
        assert mapping.__parent__.__path__ == self.__path__

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__future__ = None
        self.returncode = None
        self.stdout = None
        self.stderr = None

    def __call__(self):
        """Execute the task's program in a background process.

        A ScheduledTask may only be executed once -- subsequent
        invocations are idempotent / no-op.

        Returns the execution-initiated ScheduledTask (self).

        """
        if self.__future__ is None:
            try:
                if isinstance(exec_ := self.exec_, str):
                    cmd = plumbum.local[exec_]
                else:
                    (root, *arguments) = exec_
                    cmd = plumbum.local[root][arguments]
            except plumbum.CommandNotFound as exc:
                future = ImmediateFailure(exc)
            else:
                bound = cmd << self.param_
                future = bound.run_bg()

            self.__future__ = future

        return self

    def exception(self):
        """The in-process exception, raised by Task program
        initialization, if any.

        """
        return self.__future__.exception if isinstance(self.__future__, ImmediateFailure) else None

    def poll(self):
        """Return whether the Task program's process has exited.

        Sets the ScheduledTask's `returncode`, `stdout` and `stderr`
        when the process has exited.

        """
        if self.__future__ is None:
            return False

        if ready := self.__future__.poll():
            self.returncode = self.__future__.returncode
            self.stdout = self.__future__.stdout
            self.stderr = self.__future__.stderr

        return ready

    ready = poll

    def logs(self):
        """Parse LogRecords from `stderr`.

        Raises LogsDecodingError to indicate decoding errors when the
        encoding of a task's stderr log output is configured explicitly.
        Note, in this case, the parsed logs *may still* be retrieved
        from the exception.

        """
        if self.stderr is None:
            return None

        stream = self._iter_logs_(self.stderr)
        logs = tuple(stream)

        if stream.status.errors:
            raise LogsDecodingError(*stream.status, logs)

        return logs

    @property
    @adopt('path')
    def path_(self):
        default = super().path_
        return ScheduledTaskChainMap(*default.maps)


class ScheduledTaskChainMap(TaskChainMap):

    @at_depth('*.path')
    def result_(self, *args, **kwargs):
        return self._result_(self.__parent__.stdout, *args, **kwargs)



import itertools


class TaskProcessPool:
    """A set-size, iterable collection of tasks launched for execution
    in subprocesses.

    A TaskProcessPool may be given any size, thereby setting the maximum
    number of "slots" for concurrently-executing subprocesses. And, an
    existing pool may be expanded via method `expand`; however, pools
    may not be shrunk.

    Slots may be occupied by launched (invoked) tasks in the form of
    ScheduledTasks. Upon construction (`__init__`) and expansion
    (`expand`), an iterable of unlaunched ScheduledTasks may be
    supplied, from which available slots will be filled, with the tasks,
    launched.

    Note that -- (unlike with construction of a list in the form
    `list(iterable)`) -- only as much of the given iterable is consumed
    as necessary to fill the pool's available slots.

    Method `iter_ready` checks the pool's slotted tasks for completion,
    generating a stream of those tasks which have completed. Similar to
    the above, an iterable may be supplied of ScheduledTasks to be
    executed; these given tasks will "refill" slots emptied by completed
    tasks.

    Insofar as there are no tasks to execute, slots are occupied by
    `None`.

    """
    _fill_empty_ = itertools.repeat(None)

    def __init__(self, tasks=(), *, size=0):
        self.slots = []
        self.size = 0

        self.expand(tasks, size=size)

    def expand(self, tasks=(), *, size):
        if size < self.size:
            raise ValueError(f"cannot decrease size: {size} < {self.size}")

        if size == self.size:
            return

        tasks_fill = itertools.chain(tasks, self._fill_empty_)
        tasks_add = itertools.islice(tasks_fill, size - self.size)
        slots_add = (None if task is None else task()
                     for task in tasks_add)

        self.slots.extend(slots_add)

        self.size = int(size)

    def fill(self, tasks):
        count = 0

        refill = iter(tasks)

        for (index, slot) in enumerate(self.slots):
            if slot is None:
                try:
                    task = next(refill)
                except StopIteration:
                    break

                self.slots[index] = task()
                count += 1

        return count

    def __iter__(self):
        return iter(self.slots)

    def enumerate_tasks(self):
        # ensure index reflects actual index in slots (rather than in
        # this method's resulting iterator)
        for (index, slot) in enumerate(self):
            if slot is not None:
                yield (index, slot)

    def iter_tasks(self):
        for (_index, task) in self.enumerate_tasks():
            yield task

    def iter_ready(self, refill=None, *, clear=None):
        if not clear and clear is not None and refill is not None:
            raise TypeError(f"ambiguous argumentation: refill specified as "
                            f"{refill!r} yet clear false-y as {clear!r}")

        reservoir = None if refill is None else iter(refill)

        count = 0

        for (index, task) in self.enumerate_tasks():
            if task.ready():
                count += 1

                if reservoir:
                    task_fill = next(reservoir, None)
                    self.slots[index] = None if task_fill is None else task_fill()
                elif clear:
                    self.slots[index] = None

                yield task

        return count

    @property
    def active(self):
        try:
            next(self.iter_tasks())
        except StopIteration:
            return False
        else:
            return True

    @property
    def count(self):
        return sum(1 for _task in self.iter_tasks())



import abc
import collections

from fate.util.abstract import abstractmember


class ProxyCollection(abc.ABC):
    """Abstract mix-in of collection classes which mediate interaction
    with a proxied object (such as a list or a dict).

    Concrete subclasses must at least define class-level attribute
    `__collection_type__`, _e.g._:

        __collection_type__ = list

    """
    __collection_type__ = abstractmember()

    __slots__ = ('__collection__',)

    def __init__(self, *args, **kwargs):
        self.__collection__ = self.__collection_type__(*args, **kwargs)

    def __getitem__(self, key):
        return self.__collection__[key]

    def __len__(self):
        return len(self.__collection__)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__collection__!r})'


class ProxyMutableCollection(ProxyCollection):
    """Mutable extension to ProxyCollection."""

    __slots__ = ()

    def __setitem__(self, key, value):
        self.__collection__[key] = value

    def __delitem__(self, key):
        del self.__collection__[key]


class ProxyMapping(ProxyCollection, collections.abc.Mapping):
    """ProxyCollection providing read-only access to a dictionary."""

    __collection_type__ = dict

    __slots__ = ()

    def __iter__(self):
        yield from self.__collection__


class ProxyDict(ProxyMapping, ProxyMutableCollection, collections.abc.MutableMapping):
    """ProxyCollection providing a full-featured dictionary."""

    __slots__ = ()


class ProxySequence(ProxyCollection, collections.abc.Sequence):
    """ProxyCollection providing read-only access to a list."""

    __collection_type__ = list

    __slots__ = ()


class ProxyList(ProxySequence, ProxyMutableCollection, collections.abc.MutableSequence):
    """ProxyCollection providing a full-featured list."""

    __slots__ = ()

    def insert(self, index, value):
        self.__collection__.insert(index, value)



import itertools
import string
import time


def event_num(counter=None):
    """Generate an approximately-unique numeric identifier.

    Integer result is generated from the current epoch time in
    nanoseconds and a complementary arbitrary number to help avoid sub-
    nanosecond collision. (Note that the arbitrary complement is
    generated via counter, not randomly, to further approximate
    uniqueness.)

    This is intended to provide relatively *short* and inexpensively-
    computed unique identifiers within a single context (*e.g.* for a
    single host and application). See event_id().

    """
    time_ = time.time_ns()
    count = next(event_num.counter if counter is None else counter)
    return 100 * time_ + count


def make_cyclic_counter(size=100):
    return itertools.cycle(range(size))

event_num.counter = make_cyclic_counter()

event_num.make_cyclic_counter = make_cyclic_counter


def int_to_id(num, alphabet=(string.digits + string.ascii_letters)):
    """Translate a given integer to the base of a given alphabet."""
    if num == 0:
        return alphabet[0]

    base = len(alphabet)

    result = ''

    while num:
        (num, pos) = divmod(num, base)
        char = alphabet[pos]
        result = char + result

    return result


def event_id(counter=None):
    """Generate an approximately-unique alphanumeric identifier.

    This is intended to provide relatively *short* and inexpensively-
    computed unique identifiers within a single context (*e.g.* for a
    single host and application). See event_num().

    (Note that performance is impacted by Python implementation and
    considered in comparison to approximate alternatives including uuid,
    hashlib and/or random. This utility is intended to compute
    approximately-unique identifiers, shorter in string length than
    those generated by uuid and hashlib, and with comparable
    performance.)

    """
    return int_to_id(event_num(counter))



import abc
import itertools

from descriptors import (
    cachedclassproperty,
    classonlymethod,
)


class Resets(abc.ABC):

    def __init__(self):
        self._reset_ = []

    @cachedclassproperty
    def _reset_methods(cls):
        inherited = itertools.chain.from_iterable(
            getattr(base, '_reset_methods', ())
            for base in reversed(cls.__bases__)
        )

        own = filter(None, (getattr(obj, 'reset', None)
                            for obj in cls.__dict__.values()))

        merged = {
            reset.name: reset
            for reset in itertools.chain(inherited, own)
        }

        return tuple(merged.values())

    def _iter_pop_reset(self):
        for reset in self._reset_methods:
            args = (None,) if reset.optional else ()
            yield self.__dict__.pop(reset.name, *args)

    def reset(self):
        """Remove cached property data marked for "reset" from the
        instance.

        """
        try:
            artifact = tuple(self._iter_pop_reset())
        except KeyError as exc:
            raise ValueError(f"cannot reset: {exc} not set")

        self._reset_.append(artifact)


class ResetInfo:

    @classonlymethod
    def resets(cls, obj=None, /, **kwargs):
        reset = cls(**kwargs)

        return reset if obj is None else reset(obj)

    def __init__(self, name=None, *, optional=False):
        self.name = name
        self.optional = optional

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name!r}, optional={self.optional!r})'

    def __call__(self, obj):
        if hasattr(obj, 'reset'):
            raise ValueError("target already decorated")

        obj.reset = self

        if self.name is None:
            target = getattr(obj, '__func__', None) or obj

            if name := target.__name__:
                self.name = name
            else:
                raise TypeError("could not determine target name")

        return obj

resets = ResetInfo.resets



import collections
import os.path
import sys
import traceback
import wcwidth
from functools import partial

import loguru
from descriptors import cachedproperty

from fate.util.lazy import lazy_id

from . import encoder


class StructLogger:

    # structured record serialization function
    encode_record = encoder.dump_structured_log_record

    # serialized mapping base spec
    record_base = (
        # note: not easy to support colors when side-stepping
        # default message handling
        #
        # ('time', '<g>{time:YYYY-MM-DD HH:mm:ss.SSS}</>'),
        # ('level', '<lvl>{extra[level_lower]}</>'),

        # (serialization key, context key OR string for format_map)
        ('time', '{time:YYYY-MM-DD HH:mm:ss.SSS}'),
        ('level', 'level_lower'),
        ('event', 'event_id'),
    )

    def __init__(self, sinks, log_level=None):
        self._sinks = sinks
        self.log_level = log_level

    # callables passed to loguru

    @staticmethod
    def _sink_format_(target, level, context):
        format_ = '{extra[serialized]}\n'

        if target is sys.stderr:
            if target.isatty():
                stanza = '{level.icon}'

                # unicode icon cool but not monospace...
                if wcwidth.wcswidth(context['level'].icon) == 1:
                    stanza += ' '
            else:
                stanza = '<{extra[level_ordinal]}>'

            format_ = stanza + ' ' + format_

        if level.no <= 10:
            format_ += '{exception}'

        return format_

    @classmethod
    def _patch_record_(cls, context):
        extra = context['extra']

        # update "extra" in order to make these available to loguru format
        # string as well as to below "env"

        if exception := context['exception']:
            exc_stack = traceback.StackSummary.extract(traceback.walk_tb(exception.traceback))
            exc_frame = exc_stack[-1]
        else:
            exc_stack = exc_frame = None

        extra.update(
            level_lower=context['level'].name.lower(),
            level_ordinal=context['level'].no // 10,
            event_id=lazy_id(),
            exc_stack=exc_stack,
            exc_frame=exc_frame,
        )

        # allow reference to "extra" without subscription
        env = collections.ChainMap(context, extra)

        # build record to serialize
        record = {
            key: spec.format_map(env) if '{' in spec else env[spec]
            for (key, spec) in cls.record_base
        }

        if exc_frame and extra.get('exc_info', False):
            record.update(
                exc_line='{}:{}'.format(os.path.basename(exc_frame.filename), exc_frame.lineno),
                exc_frame=exc_frame.name,
            )

        if 'struct_extra' in extra:
            record.update(extra['struct_extra'])

        if 'struct' in extra:
            record.update(extra['struct'])

        if message := context['message']:
            if 'msg' in record:
                raise TypeError('conflict: log event specifies both text '
                                '"message" and struct "msg"')

            record['msg'] = message

        extra['serialized'] = cls.encode_record(record)

    # underlying loguru logger (lazily constructed)

    @cachedproperty
    def _base_logger(self):
        """The *unconfigured* and underlying Logger.

        Ensures (default) sink(s) only cleared once.

        """
        # NOTE: this level of loguru configuration may
        # NOTE: defeat thread-safety (tho not a current concern)

        loguru.logger.remove()  # clear default sink
        return loguru.logger

    def _clear_base_logger(self):
        """Clear the cache of the _base_logger."""
        try:
            del self._base_logger
        except AttributeError:
            pass

    @cachedproperty
    def _logger(self):
        """The *configured* and underlying Logger."""
        for sink in self._sinks:
            extra = sink.extra_

            if self.log_level:
                extra['level'] = self.log_level.upper()
            else:
                extra.setdefault('level', 'INFO')

            level_ = extra.level_

            del extra['level']

            self._add_sink(sink.target_, level_, **extra)

        return self._patched_logger()

    # helpers

    def _add_sink(self, target, level, format=None, **extra):
        if format is None:
            format = partial(self._sink_format_, target, level)

        self._base_logger.add(
            target,
            level=level.name,
            format=format,
            **extra
        )

    def _patched_logger(self):
        return self._base_logger.opt(depth=1).patch(self._patch_record_)

    # interface methods

    def safe(self, fallback_message="failed to construct logger as configured: using safe logger"):
        try:
            self._logger
        except Exception:
            # start over with base logger
            # (i.e. *ensure* default sink -- and any partially-configured sinks -- cleared)
            self._clear_base_logger()

            log_level = self.log_level.upper() if self.log_level else 'DEBUG'
            self._add_sink(sys.stderr, self._base_logger.level(log_level))

            logger = self._patched_logger()

            if fallback_message:
                logger.error(fallback_message)

            return logger
        else:
            return self

    def clone(self, **kwargs):
        instance = self.__class__(self._sinks, self.log_level)
        if logger := self.__dict__.get('_logger'):
            instance.__dict__['_logger'] = logger
        instance.__dict__.update(**kwargs)
        return instance

    def bind(self, **kwargs):
        return self.clone(_logger=self._logger.bind(**kwargs))

    def opt(self, **kwargs):
        return self.clone(_logger=self._logger.opt(**kwargs))

    def set(self, **kwargs):
        # loguru methods such as bind() *do not* merge settings;
        # rather, each call overrides the last.
        #
        # we won't be like that: set() merges its data on each invocation.
        #
        extra = self._logger._options[8]
        struct_extra = extra.get('struct_extra', {})
        data = {**struct_extra, **kwargs}
        return self.bind(struct_extra=data)

    @staticmethod
    def _make_struct(data=None, **kwargs):
        if data:
            if isinstance(data, str):
                struct = {'msg': data}
            else:
                struct = dict(data)

            struct.update(kwargs)

            return struct

        return kwargs

    def log(self, level, data=None, **kwargs):
        self._logger.log(level.upper(), '', struct=self._make_struct(data, **kwargs))

    def debug(self, data=None, **kwargs):
        self._logger.debug('', struct=self._make_struct(data, **kwargs))

    def info(self, data=None, **kwargs):
        self._logger.info('', struct=self._make_struct(data, **kwargs))

    def warning(self, data=None, **kwargs):
        self._logger.warning('', struct=self._make_struct(data, **kwargs))

    def error(self, data=None, **kwargs):
        self._logger.error('', struct=self._make_struct(data, **kwargs))

    def exception(self, data=None, **kwargs):
        self._logger.exception('', struct=self._make_struct(data, **kwargs))

    def critical(self, data=None, **kwargs):
        self._logger.critical('', struct=self._make_struct(data, **kwargs))

    def catch(self, *args, **kwargs):
        return self._logger.catch(*args, **kwargs)



import abc
import hashlib
import os
import time
import typing

from descriptors import cachedproperty, classproperty

from fate.conf import ConfBracketError
from fate.util.iteration import storeresult

from .scheduled_task import ScheduledTask
from .util.reset import resets, Resets


class TaskScheduler(Resets):
    """Abstract base class of task schedulers.

    Method `exec_tasks` must be implemented by concrete subclasses -- to
    execute tasks -- as a synchronous generator of completed tasks.
    (Task execution may be and likely should be asynchronous / pooled.)

    Invocation of a scheduler instance -- via `__call__` -- returns an
    iterator of completed tasks (derived from `exec_tasks`). Upon
    exhaustion, this iterator's attribute `info` is set to an instance
    of SchedInfo.

    Method `collect_tasks` generates tasks to be executed as a stream of
    ScheduledTask objects.

    Invocation of `collect_tasks` records the current timestamp on the
    file system and under scheduler property `time_check`. The timestamp
    of the previous "check" (retrieved from the file system) is stored
    under scheduler property `last_check`.

    The Boolean argument `reset` of methods `collect_tasks`, `__call__`,
    (and `reset`), serve to clear the above scheduler properties, such
    that a new "check" -- and a new round of task-collection -- may take
    place. Otherwise, `collect_tasks` is idempotent.

    """
    class SchedInfo(typing.NamedTuple):

        count: int
        next: float

    def __init__(self, conf, logger):
        super().__init__()
        self.conf = conf
        self.logger = logger.set(sched=self.module_short)

    @abc.abstractmethod
    def exec_tasks(self, reset=False):
        yield from ()
        return 0

    @storeresult('info')
    def __call__(self, reset=False):
        count = yield from self.exec_tasks(reset=reset)
        return self.SchedInfo(count, self.next_check)

    @classproperty
    def module_short(cls):
        (*_root, name) = cls.__module__.rsplit('.', 1)
        return name.replace('_', '-')

    @cachedproperty
    def path_check(self):
        signature = os.pathsep.join(sorted(str(conf.__path__) for conf in self.conf))
        file_hash = hashlib.md5(signature.encode()).hexdigest()
        return self.conf._prefix_.state / 'check' / file_hash

    def _check_state_(self, update=False):
        try:
            stat_result = os.stat(self.path_check)
        except FileNotFoundError:
            last_check = None
        else:
            last_check = stat_result.st_mtime

        if update:
            if not self.path_check.exists():
                self.path_check.mkdir(parents=True)
                for conf in self.conf:
                    link_path = self.path_check / conf.__path__.name
                    link_path.symlink_to(conf.__path__)

            os.utime(self.path_check, (self.time_check, self.time_check))

        return last_check

    @resets
    @cachedproperty
    def last_check(self):
        return self._check_state_(update=True)

    @resets
    @cachedproperty
    def time_check(self):
        return time.time()

    @resets
    @cachedproperty
    def _next_check_tasks_(self):
        next_check = None

        for task in self.conf.task.values():
            next_check = task.schedule_next_(
                self.time_check,             # t0
                next_check,                  # t1
                next_check,                  # default
                max_years_between_matches=1  # quit if it's that far out
            )

        return next_check

    next_max = 60 * 60 * 24 * 365  # 1 year in seconds

    @property
    def _next_check_max_(self):
        return self.time_check + self.next_max

    @property
    def next_check(self):
        return self._next_check_tasks_ or self._next_check_max_

    def _iter_tasks_due_(self):
        # bring last_check & time_check local to ensure consistent generation
        # across resets (as unlikely/discouraged as this situation might be).
        if (last_check := self.last_check) is None:
            return

        time_check = self.time_check

        for task in self.conf.task.values():
            if task.scheduled_(last_check, time_check):
                try:
                    may_schedule = task.if_
                except ConfBracketError as exc:
                    self.logger.warning(task=task.__name__, key=exc.path, msg=str(exc))
                    may_schedule = exc.evaluation

                if not may_schedule:
                    self.logger.info(task=task.__name__,
                                     msg='skipped: suppressed by if/unless condition')
                    continue

                yield ScheduledTask.schedule(task)

    def collect_tasks(self, reset=False):
        """Generate ScheduledTasks to be executed.

        If `reset` then the scheduler's current "check" is cleared, such
        that subsequently-scheduled tasks may be generated. Otherwise,
        the set of scheduled tasks is idempotent.

        """
        if reset:
            self.reset()

        yield from self._iter_tasks_due_()



"""Task logging integrated with the Fate scheduler's own logging.

Logging is provided by and configurable through `TaskLogger`.

A module-level instance of `TaskLogger` is lazily constructed and
accessible at `logger`.

Additionally, the logging methods of this module-level instance are made
dynamically available at the module level.

For example, `TaskLogger`:

    from fate.task.log import TaskLogger

    my_logger = TaskLogger(format='toml')

    my_logger.info({'ears': 'floppy'})

Module-level logger:

    from fate.task.log import logger

    logger.info({'ears': 'floppy'})

Module-level methods:

    from fate.task import log

    log.info({'ears': 'floppy'})

"""
import logging
import sys
import threading

from fate.util.format import Dumper


def __dir__():
    """Ensure "logger", and its methods, are listed, (even if it doesn't
    *yet* exist).

    """
    members = globals().keys() | _DYNAMIC_NAMES
    return sorted(members)


def __getattr__(name):
    """Lazily construct a default TaskLogger under the name "logger" and
    provide module-level access to its logging methods.

    """
    if name in _DYNAMIC_NAMES:
        context = globals()

        with threading.Lock():
            if 'logger' in context:
                logger = context['logger']
            else:
                logger = context['logger'] = TaskLogger()

        return logger if name == 'logger' else getattr(logger, name)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class TaskLogger:
    """Logging for Fate-compliant tasks."""

    def __init__(self, *,
                 end='\0',
                 file=sys.stderr,
                 format='json',
                 insert_level=False,
                 prepend_level=True):
        self.end = end
        self.file = file
        self.format = format
        self.insert_level = insert_level
        self.prepend_level = prepend_level

        try:
            self.dump = Dumper[self.format]
        except KeyError:
            raise ValueError(f'unsupported format: {format!r} not in {set(Dumper.__members__)}')

    def log(self, level, arg=None, /, **kwargs):
        """Write a record to the configured file.

        A mappping may be specified as a single positional dict argument
        and/or keyword arguments. Mappings are serialized according to
        the configured format.

        Alternatively, a string may be specified as a single positional
        argument. Keyword arguments are interpolated via format() /
        format_map().

        Examples:

            logger.log('DEBUG', transmogrified=True, output="radical")

            logger.log('WARNING', {'problems': 6}, radical=False)

            logger.log('CRITICAL', "Where have all the {plural_noun} gone?", plural_noun='cowboys')

        """
        level_code = logging._nameToLevel.get(level)

        if not level_code:
            raise ValueError(f"unrecognized log level '{level}'")

        if arg is None:
            arg = {}

        if isinstance(arg, dict):
            struct_level = {'level': level.lower()} if self.insert_level else {}

            struct = {**struct_level, **arg, **kwargs}

            message = self.dump(struct)
        elif isinstance(arg, str):
            message = arg.format_map(kwargs) if kwargs else arg
        else:
            raise TypeError('log expects at most one positional argument of '
                            f'type dict or str not {arg.__class__.__name__}')

        if self.prepend_level:
            level_ordinal = level_code // 10
            message = f'<{level_ordinal}> {message}'

        print(message, end=self.end, file=self.file)

    def critical(self, *args, **kwargs):
        """Write a CRITICAL record to the configured file.

        See log() for more information.

        """
        self.log('CRITICAL', *args, **kwargs)

    def error(self, *args, **kwargs):
        """Write an ERROR record to the configured file.

        See log() for more information.

        """
        self.log('ERROR', *args, **kwargs)

    def warning(self, *args, **kwargs):
        """Write a WARNING record to the configured file.

        See log() for more information.

        """
        self.log('WARNING', *args, **kwargs)

    warn = warning

    def info(self, *args, **kwargs):
        """Write an INFO record to the configured file.

        See log() for more information.

        """
        self.log('INFO', *args, **kwargs)

    def debug(self, *args, **kwargs):
        """Write a DEBUG record to the configured file.

        See log() for more information.

        """
        self.log('DEBUG', *args, **kwargs)


_LOGGER_NAMES = {name for name in TaskLogger.__dict__ if not name.startswith('_')}

_DYNAMIC_NAMES = {'logger'} | _LOGGER_NAMES

__all__ = {'TaskLogger'} | _DYNAMIC_NAMES



import functools

from descriptors import classonlymethod

from fate.util.sentinel import UndefinedType


class ResultIter:

    @classonlymethod
    def storeresult(cls, *c_args, **c_kwargs):
        if len(c_args) == 0 or not callable(c_args[0]):
            def decorator(generator):
                return cls.storeresult(generator, *c_args, **c_kwargs)

            return decorator

        (generator, *c_args) = c_args

        if not callable(generator):
            raise TypeError(f"'{generator.__class__.__name__}' object is not callable")

        @functools.wraps(generator)
        def wrapped(*g_args, **g_kwargs):
            iterator = generator(*g_args, **g_kwargs)
            return cls(iterator, *c_args, **c_kwargs)

        return wrapped

    def __init__(self, iterable, attr='value'):
        self.iterator = iter(iterable)
        self._name_ = attr
        self.__value__ = self._initial_ = UndefinedType()

    @property
    def done(self):
        return self.__value__ is not self._initial_

    @property
    def __value__(self):
        return getattr(self, self._name_)

    @__value__.setter
    def __value__(self, value):
        setattr(self, self._name_, value)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self.iterator)
        except StopIteration as stop:
            self.__value__ = stop.value
            raise

    def __repr__(self):
        return (f'<{self.__class__.__name__}({self.iterator!r}, {self._name_!r}): '
                f'{self.__value__!r}>')


storeresult = ResultIter.storeresult



"""Configurable support for serialization formats."""
import collections
import csv
import json
from functools import partial

import toml
import yaml
from descriptors import classproperty

from fate.util.datastructure import CallableEnum, FileFormatEnum


class ConfigurableYamlLoader(yaml.SafeLoader):
    """YAML SafeLoader supporting alternate collection classes.

    Mapping objects default to the built-in `dict`. Sequences to the
    built-in `list`. Alternate constructors may be specified upon
    initialization of the loader.

    Note: To use with the default interface, `yaml.load()`, a `partial`
    must be constructed, for example:

        yaml.load(stream, functools.partial(ConfigurableYamlLoader,
                                            dict_=NewDict))

    """
    def __init__(self, stream, dict_=None, list_=None):
        super().__init__(stream)
        self.dict_ = dict_
        self.list_ = list_

    def construct_yaml_seq(self, node):
        data = [] if self.list_ is None else self.list_()
        yield data
        data.extend(self.construct_sequence(node))

    def construct_yaml_map(self, node):
        data = {} if self.dict_ is None else self.dict_()
        yield data
        value = self.construct_mapping(node)
        data.update(value)


ConfigurableYamlLoader.add_constructor('tag:yaml.org,2002:seq',
                                       ConfigurableYamlLoader.construct_yaml_seq)

ConfigurableYamlLoader.add_constructor('tag:yaml.org,2002:map',
                                       ConfigurableYamlLoader.construct_yaml_map)


class ConfigurableTomlDecoder(toml.TomlDecoder):

    def __init__(self, _dict=dict, _list=list):
        super().__init__(_dict=_dict)
        self._list = _list

    def load_array(self, a):
        result = super().load_array(a)
        return result if type(result) == self._list else self._list(result)


class JSONEncoder(json.JSONEncoder):
    """JSON encoder supporting abstract Sequence and Mapping objects."""

    def default(self, obj):
        if isinstance(obj, collections.abc.Mapping):
            return dict(obj)

        if isinstance(obj, collections.abc.Sequence):
            return tuple(obj)

        # let base class raise TypeError
        return super().default(obj)

json_encoder = JSONEncoder()


class tag:
    """configurable decorator to set given attribute to given value"""

    _undefined_ = object()

    def __init__(self, name, value=_undefined_):
        self.name = name
        self.value = value

    def __call__(self, value_or_target):
        if self.value is self._undefined_:
            return self.__class__(self.name, value_or_target)

        setattr(value_or_target, self.name, self.value)
        return value_or_target

    def __repr__(self):
        value = 'Undefined' if self.value is self._undefined_ else repr(self.value)
        return f"{self.__class__.__name__}({self.name!r}, {value})"


raises = tag('raises')

auto = tag('auto', True)


class _NameList:

    @classproperty
    def __names__(cls):
        return [member.name for member in cls]


class _Raises:

    @property
    def raises(self):
        return getattr(self.value, 'raises', ())


class SLoader(_NameList, _Raises, FileFormatEnum, CallableEnum):

    @CallableEnum.member
    @raises(csv.Error)
    def csv(text):
        return list(csv.reader(text.splitlines()))

    @CallableEnum.member
    @auto
    @raises(json.decoder.JSONDecodeError)
    def json(text, dict_=None):
        if dict_ is None:
            return json.loads(text)

        return json.loads(text, object_hook=dict_)

    @CallableEnum.member
    @auto
    @raises(toml.decoder.TomlDecodeError)
    def toml(text, **types):
        switched = {f'_{name}'.rstrip('_'): value for (name, value) in types.items()}
        decoder = ConfigurableTomlDecoder(**switched)
        return toml.loads(text, decoder=decoder)

    @CallableEnum.member
    @auto
    @raises(yaml.error.YAMLError)
    def yaml(text, **types):
        conf = yaml.load(text, partial(ConfigurableYamlLoader, **types))
        return {} if conf is None else conf

    @classproperty
    def __auto__(cls):
        return [member for member in cls if member.auto]

    @property
    def auto(self):
        return getattr(self.value, 'auto', False)

    @classmethod
    def autoload(cls, content, format_, **types):
        if not format_:
            return (None, None)

        if format_ == 'auto' or format_ == 'mixed':
            if not content:
                return (None, None)

            for loader in cls.__auto__:
                try:
                    result = loader(content, **types)
                except loader.raises:
                    pass
                else:
                    # yaml.load treats a string literal as a valid document;
                    # however, this is probably not what's wanted.
                    if not isinstance(result, str):
                        return (result, loader)
            else:
                return (None, None)

        raise cls.NonAutoError(format_)


class NonAutoError(ValueError):
    pass

SLoader.NonAutoError = NonAutoError


class Loader(_Raises, FileFormatEnum, CallableEnum):

    @CallableEnum.member
    @raises(toml.decoder.TomlDecodeError)
    def toml(path, **types):
        with open(path) as fd:
            return SLoader.toml(fd.read(), **types)

    @CallableEnum.member
    @raises(yaml.error.YAMLError)
    def yaml(path, **types):
        with open(path) as fd:
            return SLoader.yaml(fd, **types)


class Dumper(_NameList, CallableEnum):

    @CallableEnum.member
    def json(obj):
        return json_encoder.encode(obj)

    @CallableEnum.member
    def toml(obj):
        return toml.dumps(obj)

    @CallableEnum.member
    def yaml(obj):
        return yaml.dump(obj)



import pathlib
import typing
from functools import partial

import toml

from fate.util.lazy import lstr


class TomlStrValue:

    _as_str_ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dump_funcs.update(self._dump_funcs_str_)

    @property
    def _dump_funcs_str_(self):
        return dict.fromkeys(self._as_str_, self._value_as_str_)

    def _value_as_str_(self, value):
        return self.dump_funcs[str](str(value))


class TomlTerse:

    class InlineTableFormatters(typing.NamedTuple):

        sep: typing.Iterable[str] = (', ', ' = ')
        close: typing.Iterable[str] = ('{ ', ' }')
        end: str = '\n'

    def dump_inline_table_formatted(self, section,
                                    formatters=InlineTableFormatters(),
                                    recursive_formatters=None):
        """Preserve inline table in its compact syntax instead of expanding
        into subsection.

        https://github.com/toml-lang/toml#user-content-inline-table

        Unlike the built-in `dump_inline_table`, seperators may be
        customized, for example to omit superfluous whitespace.

        Also unlike the built-in, keys are cast to str, rather than
        presumed to be this type.

        By default, configured formatting is not applied to nested
        dictionaries; these are dumped with formatting compatible to
        that of the built-in method (*not* specially "formatted").
        Specify argument `recursive_formatters=True` to apply formatting
        configured by `formatters` to nested dictionaries as well; or,
        use `recursive_formatters` to specify an alternative set of
        formatters.

        """
        if not isinstance(section, dict):
            return str(self.dump_value(section))

        (
            (sep_item, sep_pair),
            (close0, close1),
            end,
        ) = formatters

        if hasattr(recursive_formatters, '__iter__'):
            dump_value = partial(self.dump_inline_table_formatted,
                                 formatters=recursive_formatters,
                                 recursive_formatters=True)
        elif recursive_formatters:
            dump_value = partial(self.dump_inline_table_formatted,
                                 formatters=formatters,
                                 recursive_formatters=True)
        else:
            dump_value = self.dump_inline_table_formatted

        values = (
            str(key) + sep_pair + dump_value(val)
            for (key, val) in section.items()
        )
        return close0 + sep_item.join(values) + close1 + end


class TomlLoggingEncoder(TomlTerse, TomlStrValue, toml.TomlEncoder):

    _as_str_ = (lstr, pathlib.PosixPath)


toml_logging_encoder = TomlLoggingEncoder()

inline_formatters_terse = TomlLoggingEncoder.InlineTableFormatters(sep=(' ', '='),
                                                                   close=('', ''),
                                                                   end='')

nested_formatters_terse = TomlLoggingEncoder.InlineTableFormatters(sep=(' ', '='),
                                                                   close=('{', '}'),
                                                                   end='')


def dump_structured_log_record(struct):
    return toml_logging_encoder.dump_inline_table_formatted(
        struct,
        inline_formatters_terse,
        nested_formatters_terse,
    )



import enum
import functools

import fate.conf


class ExitOnError:

    def __init__(self, parser):
        self.parser = parser

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, _traceback):
        if exc_type is None:
            return

        if issubclass(exc_type, fate.conf.MultiConfError):
            paths = ', '.join(exc_value.paths)
            self.parser.exit(64, f'{self.parser.prog}: error: multiple configuration file '
                                 f'formats at overlapping paths: {paths}\n')

        if issubclass(exc_type, fate.conf.ConfSyntaxError):
            self.parser.exit(65, f'{self.parser.prog}: error: could not decode '
                                 f'{exc_value.format.upper()}: {exc_value.decode_err}\n')

        if issubclass(exc_type, fate.conf.NoConfError):
            self.parser.exit(72, f'{self.parser.prog}: error: missing '
                                 f'configuration file (tried: {exc_value})\n')

        if issubclass(exc_type, (fate.conf.ConfTypeError, fate.conf.ConfValueError)):
            self.parser.exit(78, f'{self.parser.prog}: error: {exc_value}\n')


def exit_on_error(method):
    """Decorator to apply context manager `ExitOnError` to instance
    methods of classes of type `argcmdr.Command`.

    Note: The decorator wrapper depends upon instance attribute
    `parser`. As such, affected command classes must extend the default
    `__init__` (*i.e.* `super()`); or, decorated methods must be invoked
    *after* command initialization (*i.e.* not as part of its `__init__`).

    """
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        with ExitOnError(self.parser):
            return method(self, *args, **kwargs)

    return wrapped


class CommandInterface:

    class CommandStatus(enum.Enum):
        """Status categories of task command return codes."""

        OK = 0
        Retry = 42
        Error = -1  # any other

        @classmethod
        def status(cls, code):
            """Retrieve appropriate status for given return code."""
            value = int(code)

            try:
                return cls(value)
            except ValueError:
                if value > 0:
                    return cls.Error

                raise

        def __str__(self):
            return self.name

    @property
    def conf(self):
        if (root := self.root) is None:
            # this is the root command
            # retrieve and store conf here
            try:
                conf = self.__dict__['conf']
            except KeyError:
                conf = self.__dict__['conf'] = self.args.__conf__ or fate.conf.get()

            return conf

        # defer to root
        return root.conf

    @property
    def exit_on_error(self):
        return ExitOnError(self.parser)

    @staticmethod
    def write_result(path, contents):
        if path.exists():
            raise FileExistsError(path)

        if not path.parent.exists():
            try:
                path.parent.mkdir(parents=True)
            except NotADirectoryError:
                pass

        if not path.parent.is_dir():
            raise NotADirectoryError(20, 'Not a directory', path.parent)

        with path.open('w') as fd:
            print(contents, end='', file=fd)



import collections
import functools
import logging
import pathlib
import re
from datetime import datetime
from enum import IntEnum

import croniter
import jinja2
from descriptors import classproperty

from fate.util.compat.os import cpu_count
from fate.util.format import Dumper, SLoader
from fate.util.iteration import storeresult
from fate.util.sentinel import Undefined

from fate.util.datastructure import (
    adopt,
    at_depth,
    StrEnum,
)

from .. import template

from ..error import (
    ConfBracketError,
    ConfTypeError,
    ConfValueError,
    ResultEncodingError,
)

from .base import (
    ConfChain,
    ConfDict,
    ConfType,
)


class TaskConfType(ConfType):
    """Generic interface applied to data deserialized from task
    configuration files.

    """
    _Dumper = Dumper
    _Loader = SLoader

    _LogRecord = collections.namedtuple('LogRecord', ('level', 'record'))

    _LogDecodingStatus = collections.namedtuple('LogDecodingStatus', ('format', 'errors'))

    class _DefaultFormat(StrEnum):

        param = 'json'
        log = 'auto'
        result = 'auto'

    class _DefaultScheduling(IntEnum):

        tenancy = (2 * cpu_count() - 1)

    class _ShellExecutable(StrEnum):

        sh = 'sh'
        bash = 'bash'
        python = 'python'

        @classproperty
        def default(cls):
            return cls.sh

        @classproperty
        def options(cls):
            return [str(member) for member in cls]

    class _ScheduleError(StrEnum):

        missing = "{task_name}: schedule requires cron expression like '* * * * *'"
        bad = missing + " not {schedule!r}"

        def self_format(self, task, schedule=None):
            return self.format(task_name=task.__name__, schedule=schedule)

    _serializer_error = ('{conf_path}: unsupported serialization format: {format_!r} '
                         f"(select from: {_Loader.__names__})")

    def __repr__(self):
        repr_ = super().__repr__()

        if (path := self.__path__) is None:
            return repr_

        return f"<{self.__class__.__name__} @ {path}: {repr_}>"

    @property
    def __default__(self):
        return self.__root__.__other__.default

    @at_depth(0)
    @property
    def _croniter(self):
        # cover for croniter_range not directly supporting hash_id
        return functools.partial(croniter.croniter, hash_id=self.__name__)

    @at_depth(0)
    def schedule_iter_(self, t0, t1=None, /, *, max_years_between_matches=None):
        try:
            schedule = self['schedule']
        except KeyError:
            raise ConfTypeError(self._ScheduleError.missing.self_format(self))

        if not isinstance(schedule, str):
            raise ConfTypeError(self._ScheduleError.bad.self_format(self, schedule))

        runs = (self._croniter(schedule, t0,
                               max_years_between_matches=max_years_between_matches)
                if t1 is None
                else croniter.croniter_range(t0, t1, schedule, _croniter=self._croniter))

        try:
            yield from runs
        except croniter.CroniterBadCronError:
            raise ConfValueError(self._ScheduleError.bad.self_format(self, schedule))

    @at_depth(0)
    def schedule_next_(self, *args, **kwargs):
        arglen = len(args)

        if 1 <= arglen <= 2:
            (times, default) = (args, Undefined)
        elif arglen == 3:
            (*times, default) = args
        else:
            raise TypeError(f"expected 2 or 3 arguments got {arglen}")

        series = self.schedule_iter_(*times, **kwargs)
        return next(series) if default is Undefined else next(series, default)

    @at_depth(0)
    def scheduled_(self, *times):
        try:
            next(self.schedule_iter_(*times))
        except StopIteration:
            return False
        else:
            return True

    @at_depth(0)
    @property
    @adopt('scheduling')
    def scheduling_(self):
        return TaskChainMap(
            self.get('scheduling', {}),
            self.__default__.get('scheduling', {}),
            self._DefaultScheduling.__members__,
        )

    @at_depth(0)
    @property
    def exec_(self):
        # check for superfluous/conflicting argumentation
        keys = ('exec', 'shell', 'command')
        if sum(key in self for key in keys) > 1:
            raise ConfTypeError(f"{self.__name__}: ambiguous task configuration: "
                                f"specify 0 or 1 of keys: {keys!r}")

        if 'exec' in self:
            try:
                return template.render_complex(self['exec'])
            except TypeError:
                raise ConfTypeError(f'{self.__name__}: "exec" requires string or list of strings '
                                    f"not: {self['exec']!r}")
            except jinja2.TemplateError as exc:
                raise ConfValueError(f"{exc.__class__.__name__} @ {self.__path__}.exec: {exc}")

        if 'shell' in self:
            if isinstance(self.shell, collections.abc.Mapping) and 'executable' in self.shell:
                executable = self.shell.executable

                if not isinstance(executable, str):
                    raise ConfTypeError(f'{self.__name__}.shell: "executable" expected '
                                        f'string not: {executable!r}')

                if executable not in self._ShellExecutable:
                    raise ConfValueError(f'{self.__name__}.shell: "executable" must be '
                                         f'one of: {tuple(self._ShellExecutable.options)!r}')
            else:
                executable = self._ShellExecutable.default

            script = (self.shell.get('script') if isinstance(self.shell, collections.abc.Mapping)
                      else self.shell)

            if not isinstance(script, str):
                raise ConfTypeError(
                    f'{self.__name__}: "shell" expected string input '
                    f"for '{executable}' command indicated either by a simple string "
                    f"(for '{self._ShellExecutable.default}') or a mapping of string "
                    f"input ('script') and optional base command "
                    f"('executable': {set(self._ShellExecutable.options)!r})"
                )

            if not script:
                raise ConfValueError(
                    f'{self.__name__}: "shell" expected string input '
                    f"for '{executable}' command indicated either by a simple string "
                    f"(for '{self._ShellExecutable.default}') or a mapping of string "
                    f"input ('script') and optional base command "
                    f"('executable': {set(self._ShellExecutable.options)!r})"
                )

            try:
                return (executable, '-c', template.render_str(script))
            except jinja2.TemplateError as exc:
                raise ConfValueError(f"{exc.__class__.__name__} @ {self.__path__}.shell: {exc}")

        command = self['command'] if 'command' in self else self.__name__

        if not isinstance(command, str):
            raise ConfTypeError(f'{self.__name__}: "command" requires string not: {command!r}')

        return f'{self.__lib__}-{command}'

    @at_depth(0)
    @property
    def if_(self):
        if 'if' in self:
            if 'unless' in self:
                raise ConfTypeError(f"{self.__name__}: ambiguous configuration: "
                                    "specify either task 'if' or 'unless' not both")

            (key, negate) = ('if', False)
        elif 'unless' in self:
            (key, negate) = ('unless', True)
        else:
            return True

        expression = self[key]

        if not isinstance(expression, str):
            raise ConfTypeError(f'{self.__name__}: "{key}" requires expression '
                                f"string not: {expression!r}")

        bracket_match = template.variable_pattern.fullmatch(expression.strip())

        target = bracket_match['expr'] if bracket_match else expression

        try:
            evaluation = template.eval_expr(target)
        except jinja2.TemplateError as exc:
            raise ConfValueError(f"{exc.__class__.__name__} @ {self.__path__}.{key}: {exc}")

        predicate = not evaluation if negate else bool(evaluation)

        if bracket_match:
            path = self.__get_path__(key)[1:]
            raise ConfBracketError(path, predicate)

        return predicate

    @at_depth(0)
    @property
    def format_(self):
        return collections.ChainMap(
            self.get('format', {}),
            self.__default__.get('format', {}),
            self._DefaultFormat.__members__,
        )

    @at_depth(0)
    @property
    @adopt('path')
    def path_(self):
        return TaskChainMap(
            self.get('path', {}),
            self.__default__.get('path', {}),
            {'result': self._prefix_.data / 'result'},
        )

    @at_depth(0)
    @property
    def param_(self):
        param = self.get('param', {})

        if isinstance(param, str):
            return param

        format_ = self.format_['param']

        try:
            dumper = self._Dumper[format_]
        except KeyError:
            raise ConfValueError(
                f'{self.__name__}: unsupported serialization format: '
                f"{format_!r} (select from: {self._Dumper.__names__})"
            )
        else:
            return dumper(param)

    _task_log_pattern = re.compile(r'<([1-5])> *(.*)')

    @staticmethod
    def _normalize_log_level(value):
        if isinstance(value, int):
            if value < 10:
                value *= 10

            return logging._levelToName.get(value)

        if isinstance(value, str):
            value = value.upper()

            return value if value in logging._nameToLevel else None

        raise TypeError("log level to normalize may be of type int or str "
                        f"not {value.__class__.__name__}")

    @at_depth(0)
    @storeresult('status')
    def _iter_logs_(self, stderr):
        format_ = self.format_['log']

        status = self._LogDecodingStatus(format_, [])

        for log_line in stderr.split('\0'):
            if not log_line:
                continue

            if level_match := self._task_log_pattern.fullmatch(log_line):
                (record_code, record_text) = level_match.groups()
                record_level = self._normalize_log_level(int(record_code))
            else:
                (record_level, record_text) = (None, log_line)

            try:
                (record_struct, _loader) = self._Loader.autoload(record_text, format_)
            except self._Loader.NonAutoError:
                try:
                    loader = self._Loader[format_]
                except KeyError:
                    raise ConfValueError(
                        self._serializer_error.format(
                            conf_path=f'{self.__name__}.format.log',
                            format_=format_,
                        )
                    )

                try:
                    record_struct = loader(record_text)
                except loader.raises as exc:
                    status.errors.append(exc)
                    record_struct = None

            struct_level = record_struct.get('level') if isinstance(record_struct, dict) else None

            if struct_level:
                if isinstance(struct_level, (int, str)):
                    struct_level = self._normalize_log_level(struct_level)
                else:
                    struct_level = None

            yield self._LogRecord(
                level=(record_level or struct_level or 'INFO'),
                record=(record_struct if isinstance(record_struct, dict) else record_text),
            )

        return status

    @at_depth('*.path')
    def _result_(self, stdout, dt=None):
        result_spec = self.result

        if not result_spec:
            # empty path.result synonymous with /dev/null: quit
            return None

        format_ = self.__parent__.format_['result']

        if format_ == 'auto' and not stdout:
            # auto (unlike mixed) suppresses empty results: quit
            return None

        if dt is None:
            dt = datetime.now()

        stamp = dt.timestamp()
        datestr = dt.strftime('%Y%m%dT%H%M%S')

        result_path = (result_spec if isinstance(result_spec, pathlib.Path)
                       else pathlib.Path(result_spec))

        identifier = result_path / f'result-{stamp:.0f}-{datestr}-{self.__parent__.__name__}'

        try:
            (_struct, loader) = self._Loader.autoload(stdout, format_)
        except self._Loader.NonAutoError:
            if isinstance(format_, str) or not isinstance(format_, collections.abc.Iterable):
                formats = (format_,)
            else:
                formats = format_

            errors = []

            for format1 in formats:
                try:
                    loader = self._Loader[format1]
                except KeyError:
                    raise ConfValueError(
                        self._serializer_error.format(
                            conf_path=f'{self.__parent__.__name__}.format.result',
                            format_=format1,
                        )
                    )

                try:
                    loader(stdout)
                except loader.raises as exc:
                    errors.append(exc)
                else:
                    break
            else:
                raise ResultEncodingError(format_, errors, identifier)

        return identifier.with_suffix(loader.suffix) if loader else identifier


class TaskConfDict(TaskConfType, ConfDict):
    """Mapping of data deserialized from task configuration files."""


class TaskChainMap(TaskConfType, ConfChain):
    """ChainMap of task configuration with fallback to system-level and
    library-level defaults.

    """



"""Inference, and environment variable specification, of relevant
filesystem paths.

"""
import enum
import functools
import operator
import os
import re
import sys
from dataclasses import dataclass
from pathlib import Path

from descriptors import cachedproperty, classonlymethod, classproperty

from fate.util.compat.path import is_relative_to


class PrefixProfile(enum.Flag):

    #
    # system: library was installed under a system (root-writable) path
    #
    # (if not set, library was installed under some user path.)
    #
    system = enum.auto()

    #
    # isolated: library was installed into a python virtual environment
    #           for the purpose of isolation
    #
    # (note: this flag may not be set for some virtual environments,
    # which seek only to isolate the library's requirements, but not
    # the tool itself -- e.g., pipx)
    #
    isolated = enum.auto()

    @classproperty
    def empty(cls):
        return cls(0)

    @classonlymethod
    def infer(cls, lib):
        """Compose profile flags appropriate to the installation context.

        Inference is overridden by the process environment variable:

            {LIB}_PREFIX_PROFILE={system,isolated,empty}

        The value of this variable may be a singular profile name, or a
        list of these, delimited by either "," (comma) or "|" (pipe).

        The special value `empty` may be specified to indicate the empty
        flag, (*i.e.*, non-system, non-isolated, or the "user-global"
        profile).

        Note: Any other non-existant profile names specified by the
        environment variable are ignored.

        """
        #
        # check for profile specified by environment variable
        #

        environ_spec = os.getenv(f'{lib}_PREFIX_PROFILE'.upper(), '')

        # environ-given names may be delimited by comma or pipe
        environ_names = re.findall(r'[^ ,|]+', environ_spec)

        # unrecognized environ-given names are simply ignored
        environ_profiles = [cls[name] for name in environ_names if name in cls.__members__]

        if 'empty' in environ_names:
            environ_profiles.append(cls.empty)

        if environ_profiles:
            return functools.reduce(operator.or_, environ_profiles)

        #
        # infer from installation
        #

        # compat: Python <3.9
        user_installation = is_relative_to(Path(__file__), Path.home().parent)

        # module either installed under a user home directory
        # (and so will use XDG_CONFIG_HOME, etc.)
        # OR appears global
        # (and so will install global)
        location_profile = cls.empty if user_installation else cls.system

        if (
            # using system python
            sys.prefix == sys.base_prefix

            # using tox venv: treat as non-isolated
            or 'pipx' in Path(sys.prefix).parts
        ):
            return location_profile

        # looks isolated (a venv)
        # (and so will construct path from `sys.prefix`)
        return location_profile | cls.isolated


def path(method):
    """Wrap a Path-returning method such that:

    * it may be overridden by an environment variable

    * its result (if any) is given a leaf directory named for the
      PrefixPath.lib attribute

    """
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        if override := os.getenv(f'{self.lib}_PREFIX_{method.__name__}'.upper()):
            return Path(override).absolute()

        method_path = method(self, *args, **kwargs)

        return method_path and method_path / self.lib

    return wrapped


def path_property(prop):
    """Wrap a `Path`-returning method as a `path` and as a
    `cachedproperty`.

    See: `path`.

    """
    return cachedproperty(path(prop))

path.property = path_property


@dataclass
class PrefixPaths:
    """Path prefixes appropriate to the environment.

    Overrides to inference and defaults are retrieved from the
    process environment variables:

        {LIB}_PREFIX_PROFILE={system,isolated}

        {LIB}_PREFIX_{FIELD}=path

    """
    lib: str
    profile: PrefixProfile

    @classonlymethod
    def infer(cls, lib):
        profile = PrefixProfile.infer(lib)
        return cls(lib, profile)

    @path.property
    def conf(self):
        """library configuration"""
        if PrefixProfile.system in self.profile:
            return Path(os.sep) / 'etc'

        if PrefixProfile.isolated in self.profile:
            return Path(sys.prefix)

        if xdg_config := os.getenv('XDG_CONFIG_HOME'):
            return Path(xdg_config)

        return Path.home() / '.config'

    @path.property
    def data(self):
        """results directory (default)"""
        if PrefixProfile.system in self.profile:
            return Path(os.sep) / 'var' / 'log'

        if PrefixProfile.isolated in self.profile:
            return Path(sys.prefix)

        if xdg_data := os.getenv('XDG_DATA_HOME'):
            return Path(xdg_data)

        return Path.home() / '.local' / 'share'

    @path.property
    def state(self):
        """library (retry records) and task state"""
        if PrefixProfile.system in self.profile:
            return Path(os.sep) / 'var' / 'lib'

        if PrefixProfile.isolated in self.profile:
            return Path(sys.prefix)

        if xdg_state := os.getenv('XDG_STATE_HOME'):
            return Path(xdg_state)

        return Path.home() / '.local' / 'state'

    @path.property
    def run(self):
        """run (lock) files"""
        if PrefixProfile.system in self.profile:
            return Path(os.sep) / 'run'

        if PrefixProfile.isolated in self.profile:
            return Path(sys.prefix)

        if xdg_runtime := os.getenv('XDG_RUNTIME_DIR'):
            return Path(xdg_runtime)

        return Path.home() / '.local' / 'run'

    @path
    def completions(self, shell_name, force_system=None):
        """shell completion files"""
        dir_name = 'bash-completion' if shell_name == 'bash' else shell_name

        if force_system or (force_system is None and PrefixProfile.system in self.profile):
            return Path(os.sep) / 'usr' / 'share' / dir_name / 'completions'

        # completions must ignore user virtual env (and shouldn't matter)

        data_path = (Path(xdg_data) if (xdg_data := os.getenv('XDG_DATA_HOME'))
                     else Path.home() / '.local' / 'share')

        return data_path / dir_name / 'completions'



import argparse
import enum
import functools
import operator
import os
import pathlib


class ChoiceMapping(argparse.Action):
    """Argparse Action to interpret the `choices` argument as a
    mapping of user-specified choices values to the resulting option
    values.

    """
    def __init__(self, *args, choices, **kwargs):
        super().__init__(*args, choices=choices.keys(), **kwargs)
        self.mapping = choices

    def __call__(self, parser, namespace, value, option_string=None):
        setattr(namespace, self.dest, self.mapping[value])


def access_parent(path):
    access_target = path

    # Path.exists() calls stat() which can raise PermissionError (prematurely)
    while not os.path.exists(access_target) and access_target.name != '':
        access_target = access_target.parent

    return access_target


class PathAccess:
    """Argparse type ensuring filesystem access to given path argument.

    An instance of pathlib.Path is returned.

    """
    class Access(enum.IntFlag):

        r = os.R_OK
        w = os.W_OK

        @property
        def mode(self):
            #
            # note: enum has various internal means of determining contents of
            # a composite flag until proper instance iteration lands in 3.11 or
            # so.
            #
            # rather than worrying about that, here we determine contained names
            # manually. in 3.11 should be even simpler.
            #
            return ''.join(member.name for member in self.__class__ if member in self)

        def ok(self, path):
            return os.access(path, self)

    class PathAccessError(argparse.ArgumentTypeError):
        """Subclass of ArgumentTypeError raised when path permissions
        do not match specified mode.

        """

    def __init__(self, mode, parents=False):
        if isinstance(mode, str):
            self.access = functools.reduce(operator.or_, (self.Access[part] for part in mode))
        elif isinstance(mode, int):
            self.access = self.Access(mode)
        elif isinstance(mode, self.Access):
            self.access = mode
        else:
            raise TypeError('expected access mode of type str, int or Access '
                            'not ' + mode.__class__.__name__)

        self.parents = parents

    def __call__(self, value):
        path = pathlib.Path(value)

        access_target = access_parent(path) if self.parents else path

        if not self.access.ok(access_target):
            raise self.PathAccessError("failed to access path with mode "
                                       f"{self.access.mode}: {path}")

        return path


class PathTypeError(argparse.ArgumentTypeError):
    """Subclass of ArgumentTypeError raised for path of incorrect type."""


class FileAccess(PathAccess):

    PathTypeError = PathTypeError

    def __call__(self, value):
        path = super().__call__(value)

        if self.parents and not path.exists():
            if not access_parent(path).is_dir():
                raise self.PathTypeError(f"path inaccessible: {path}")
        elif not path.is_file():
            raise self.PathTypeError(f"path must be file: {path}")

        return path


class DirAccess(PathAccess):

    PathTypeError = PathTypeError

    def __call__(self, value):
        path = super().__call__(value)

        if self.parents and not path.exists():
            if not access_parent(path).is_dir():
                raise self.PathTypeError(f"path inaccessible: {path}")
        elif not path.is_dir():
            raise self.PathTypeError(f"path must be directory: {path}")

        return path



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fate',
 'fate.cli',
 'fate.cli.base',
 'fate.cli.command',
 'fate.conf',
 'fate.conf.base',
 'fate.conf.types',
 'fate.sched',
 'fate.sched.base',
 'fate.sched.base.util',
 'fate.task',
 'fate.util',
 'fate.util.compat',
 'fate.util.datastructure',
 'fate.util.log']

package_data = \
{'': ['*'], 'fate.cli': ['include/banner/*'], 'fate.conf': ['include/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'argcmdr>=1.0.0,<2.0.0',
 'argcomplete>=2.0,<3.0',
 'croniter>=1.3.5,<2.0.0',
 'loguru>=0.6.0,<0.7.0',
 'pyyaml>=6.0,<7.0',
 'schema>=0.7.5,<0.8.0',
 'toml>=0.10.2,<0.11.0',
 'wcwidth>=0.2.5,<0.3.0']

extras_require = \
{':python_version == "3.8"': ['importlib-resources==5.0']}

entry_points = \
{'console_scripts': ['fate = fate:main',
                     'fated = fate:daemon',
                     'fates = fate:serve']}

setup_kwargs = {
    'name': 'fate-scheduler',
    'version': '0.1.0rc7',
    'description': 'The operating system-level command scheduler and manager.',
    'long_description': None,
    'author': 'Jesse London',
    'author_email': 'jesselondon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/chicago-cdac/fate',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import multiprocessing
import os


COMPAT_INTERVAL = 0.02


def cpu_count():
    try:
        # glibc-only
        sched_getaffinity = os.sched_getaffinity
    except AttributeError:
        # Note: this *may* be inaccurate in some shared environs
        return multiprocessing.cpu_count()
    else:
        return len(sched_getaffinity(0))


def get_interval(default=COMPAT_INTERVAL):
    try:
        sched_rr_get_interval = os.sched_rr_get_interval
    except AttributeError:
        # unsupported (e.g. darwin)
        return default
    else:
        return sched_rr_get_interval(0)



from fate.util.datastructure import (
    AttributeChainMap,
    AttributeProxyDict,
    DecoratedNestedConf,
    ProxyList,
)


class ConfType(DecoratedNestedConf):
    """Generic base for configuration classes.

    This base does not provide the configuration's container class.

    """
    @property
    def __lib__(self):
        return self.__root__.__lib__

    @property
    def _prefix_(self):
        return self.__root__._prefix_


"""Configuration container base classes.

These do *not* extend the classes provided by datastructure. Rather,
these are set as they are for clarity and symmetry.

"""
ConfChain = AttributeChainMap

ConfDict = AttributeProxyDict

ConfList = ProxyList


class ConfMapping(ConfType, ConfDict):
    """Default base for configuration mapping classes."""



class ConfError(Exception):
    pass


class ConfSyntaxError(ConfError):

    def __init__(self, format_, decode_err):
        super().__init__(format_, decode_err)
        self.format = format_
        self.decode_err = decode_err


class ConfTypeError(TypeError, ConfError):
    pass


class ConfValueError(ValueError, ConfError):
    pass


class NoConfError(LookupError, ConfError):
    pass


class MultiConfError(ValueError, ConfError):

    @property
    def paths(self):
        return [str(path) for path in self.args]


class LogsDecodingError(ValueError, ConfError):

    def __init__(self, format_, errors, logs):
        super().__init__(format_, errors, logs)
        self.format = format_
        self.errors = errors
        self.logs = logs


class ResultEncodingError(ValueError, ConfError):

    def __init__(self, format_, errors, identifier):
        super().__init__(format_, errors, identifier)
        self.format = format_
        self.errors = errors
        self.identifier = identifier


class ConfBracketError(ValueError, ConfError):

    message = ('expression SHOULD NOT be enclosed by brackets '
               'outside of interpolation context')

    def __init__(self, path, evaluation):
        super().__init__(path, evaluation)
        self.path = path
        self.evaluation = evaluation

    def __str__(self):
        return self.message



from functools import partial

import argcmdr

import fate.cli.command
from fate.cli.base import Main


def extend_parser(parser, conf=None, banner_path=None, entry_points=None):
    parser.set_defaults(
        __conf__=conf,
        __banner_path__=banner_path,
        __entry_points__=entry_points,
    )


def entrypoint(root, **settings):
    # auto-discover nested commands
    argcmdr.init_package(
        fate.cli.command.__path__,
        fate.cli.command.__name__,
    )

    if isinstance(root, str):
        # lazily look up nested command signature
        names = root.split('.')
        root = fate.cli.command
        for name in names:
            root = getattr(root, name)

    argcmdr.main(root, extend_parser=partial(extend_parser, **settings))


main = partial(entrypoint, Main)

daemon = partial(entrypoint, 'control.Daemon')

serve = partial(entrypoint, 'control.Serve')



import re
import typing

from descriptors import cachedproperty, classonlymethod

from fate.util.datastructure import NamedTupleEnum, StrEnum

from ..path import PrefixPaths

from ..types import (
    DefaultConf,
    DefaultConfDict,
    DefaultConfList,
    TaskConfDict,
)

from .conf import Conf


class ConfSpec(typing.NamedTuple):

    name: str
    filename: typing.Optional[str] = None
    types: typing.Optional[dict] = None
    conf: typing.Optional[type] = Conf


class BuiltinConfSpec(typing.NamedTuple):

    # where in package to find built-in configuration files (with
    # comments, etc.) for conf initialization (copying into host), and
    # even for fallback in case of non-initialization (missing
    # configuration)
    #
    # (may be overridden with import path to *any* package)
    #
    path: str = 'fate.conf.include'

    # whether to permit falling back to built-in configuration in case
    # of missing host configuration
    #
    fallback: bool = False

    @classonlymethod
    def _from_group(cls, **kwargs):
        proper = {}

        for (name, value) in kwargs.items():
            if match := re.fullmatch(r'builtin_(.+)', name):
                proper[match[1]] = value
            else:
                raise TypeError(f"unexpected keyword argument {name!r}")

        return cls(**proper)


class ConfGroup:
    """Namespaced collection of Conf objects."""

    class _Spec(ConfSpec, NamedTupleEnum):

        task = ConfSpec('task', types={'dict': TaskConfDict})
        default = ConfSpec('default', conf=DefaultConf, types={'dict': DefaultConfDict,
                                                               'list': DefaultConfList})

    class _Default(StrEnum):

        lib = 'fate'

    def __init__(self, *specs, lib=None, **builtin_spec):
        self._lib_ = lib or self._Default.lib  # disallow empty

        self._builtin_ = BuiltinConfSpec._from_group(**builtin_spec)

        data = dict(self._iter_conf_(specs))
        self.__dict__.update(data)
        self._names_ = tuple(data)

        self._link_conf_()

    @cachedproperty
    def _prefix_(self):
        return PrefixPaths.infer(self._lib_)

    def _iter_conf_(self, specs):
        for spec in (specs or self._Spec):
            if isinstance(spec, str):
                spec = (spec, None, None, Conf)

            (conf_name, file_name, types, conf_type) = spec

            yield (
                conf_name,
                conf_type(
                    conf_name,
                    self._lib_,
                    self._builtin_,
                    self._prefix_,
                    file_name,
                    types,
                )
            )

    def _link_conf_(self):
        for name0 in self._names_:
            conf0 = getattr(self, name0)

            for name1 in self._names_:
                if name1 == name0:
                    continue

                conf1 = getattr(self, name1)
                setattr(conf0.__other__, name1, conf1)

    def __iter__(self):
        for name in self._names_:
            yield getattr(self, name)

    def __repr__(self):
        return f'<{self.__class__.__name__} [%s]>' % ', '.join(self._names_)



import abc
import functools

from .access import nomap
from .collection import ProxyMapping


class LazyLoadMap(abc.ABC):
    """Abstract mix-in for mapping classes to load data lazily.

    Only upon look-up is the mapping updated with the result of abstract
    method `__getdata__`.

    To customize how the mapping is updated, override `__setdata__`.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._loaded_ = False

    @abc.abstractmethod
    def __getdata__(self):
        return {}

    def __setdata__(self, data):
        self.update(data)

    def _load_(self):
        data = self.__getdata__()
        self.__setdata__(data)
        self._loaded_ = True

    def _loadif_(self):
        if not self._loaded_:
            self._load_()

    def __getitem__(self, key):
        self._loadif_()
        return super().__getitem__(key)

    def __repr__(self):
        self._loadif_()
        return super().__repr__()

    def __iter__(self):
        self._loadif_()
        return super().__iter__()

    def __len__(self):
        self._loadif_()
        return super().__len__()


class LazyLoadProxyMapping(LazyLoadMap, ProxyMapping):
    """Abstract base class for a lazy-loading ProxyMapping.

    Only upon look-up is the mapping updated with the result of abstract
    method `__getdata__`.

    """
    def __setdata__(self, data):
        self.__collection__.update(data)


class LoadAttributeError(Exception):
    """Stand-in for `AttributeError`.

    Does *not* extend `AttributeError`, such that descriptors
    responsible for lazily loading map data may raise it without
    triggering `__getattr__`.

    """


loads = functools.partial(nomap, exc_class=LoadAttributeError)



import functools

from .base import ConfGroup
from .error import (  # noqa: F401
    ConfBracketError,
    ConfSyntaxError,
    ConfTypeError,
    ConfValueError,
    LogsDecodingError,
    MultiConfError,
    NoConfError,
    ResultEncodingError,
)


spec = ConfGroup._Spec


@functools.wraps(ConfGroup, assigned=('__doc__', '__annotations__'), updated=())
def get(*args, **kwargs):
    return ConfGroup(*args, **kwargs)
